from kclvm.kcl.ast import BinOp, CmpOp, UnaryOp, AugOp, ExprContext

from kclvm.vm.code import Opcode
from kclvm.compiler.build.symtable import SymbolScope


CMP_OP_MAPPING = {
    CmpOp.Eq: Opcode.COMPARE_EQUAL_TO,
    CmpOp.NotEq: Opcode.COMPARE_NOT_EQUAL_TO,
    CmpOp.Lt: Opcode.COMPARE_LESS_THAN,
    CmpOp.LtE: Opcode.COMPARE_LESS_THAN_OR_EQUAL_TO,
    CmpOp.Gt: Opcode.COMPARE_GREATER_THAN,
    CmpOp.GtE: Opcode.COMPARE_GREATER_THAN_OR_EQUAL_TO,
    CmpOp.Is: Opcode.COMPARE_IS,
    CmpOp.IsNot: Opcode.COMPARE_IS_NOT,
    CmpOp.In: Opcode.COMPARE_IN,
    CmpOp.NotIn: Opcode.COMPARE_NOT_IN,
    CmpOp.Not: Opcode.COMPARE_IS_NOT,  # 'not' => 'is not'
}


BIN_OP_MAPPING = {
    **CMP_OP_MAPPING,
    BinOp.Add: Opcode.BINARY_ADD,
    BinOp.Sub: Opcode.BINARY_SUBTRACT,
    BinOp.Mul: Opcode.BINARY_MULTIPLY,
    BinOp.Div: Opcode.BINARY_TRUE_DIVIDE,
    BinOp.Mod: Opcode.BINARY_MODULO,
    BinOp.Pow: Opcode.BINARY_POWER,
    BinOp.LShift: Opcode.BINARY_LSHIFT,
    BinOp.RShift: Opcode.BINARY_RSHIFT,
    BinOp.BitOr: Opcode.BINARY_OR,
    BinOp.BitXor: Opcode.BINARY_XOR,
    BinOp.BitAnd: Opcode.BINARY_AND,
    BinOp.FloorDiv: Opcode.BINARY_FLOOR_DIVIDE,
    BinOp.And: Opcode.BINARY_LOGIC_AND,
    BinOp.Or: Opcode.BINARY_LOGIC_OR,
    BinOp.As: Opcode.MEMBER_SHIP_AS,
}

UNARY_OP_MAPPING = {
    UnaryOp.Invert: Opcode.UNARY_INVERT,
    UnaryOp.Not: Opcode.UNARY_NOT,
    UnaryOp.UAdd: Opcode.UNARY_POSITIVE,
    UnaryOp.USub: Opcode.UNARY_NEGATIVE,
}

ARG_OP_MAPPING = {
    AugOp.Add: Opcode.INPLACE_ADD,
    AugOp.Sub: Opcode.INPLACE_SUBTRACT,
    AugOp.Mul: Opcode.INPLACE_MULTIPLY,
    AugOp.Div: Opcode.INPLACE_TRUE_DIVIDE,
    AugOp.Mod: Opcode.INPLACE_MODULO,
    AugOp.Pow: Opcode.INPLACE_POWER,
    AugOp.LShift: Opcode.INPLACE_LSHIFT,
    AugOp.RShift: Opcode.INPLACE_RSHIFT,
    AugOp.BitOr: Opcode.INPLACE_OR,
    AugOp.BitXor: Opcode.INPLACE_XOR,
    AugOp.BitAnd: Opcode.INPLACE_AND,
    AugOp.FloorDiv: Opcode.INPLACE_FLOOR_DIVIDE,
}

EXPR_OP_MAPPING = {
    ExprContext.AUGLOAD: Opcode.LOAD_ATTR,
    ExprContext.LOAD: Opcode.LOAD_ATTR,
    ExprContext.AUGSTORE: Opcode.STORE_ATTR,
    ExprContext.STORE: Opcode.STORE_ATTR,
    ExprContext.DEL: Opcode.DELETE_ATTR,
}


SUBSCR_OP_MAPPING = {
    ExprContext.AUGLOAD: [Opcode.DUP_TOP_TWO, Opcode.BINARY_SUBSCR],
    ExprContext.LOAD: [Opcode.BINARY_SUBSCR],
    ExprContext.AUGSTORE: [Opcode.ROT_THREE, Opcode.STORE_SUBSCR],
    ExprContext.STORE: [Opcode.STORE_SUBSCR],
    ExprContext.DEL: [Opcode.DELETE_SUBSCR],
}

SYMBOL_SCOPE_LOAD_OP_MAPPING = {
    SymbolScope.BUILT_IN: Opcode.LOAD_BUILT_IN,
    SymbolScope.LOCAL: Opcode.LOAD_LOCAL,
    SymbolScope.GLOBAL: Opcode.LOAD_NAME,
    SymbolScope.FREE: Opcode.LOAD_FREE,
    SymbolScope.INTERNAL: Opcode.LOAD_NAME,
}

SYMBOL_SCOPE_STORE_OP_MAPPING = {
    SymbolScope.GLOBAL: Opcode.STORE_GLOBAL,
    SymbolScope.LOCAL: Opcode.STORE_LOCAL,
    SymbolScope.FREE: Opcode.STORE_FREE,
}


class CompilerInternalErrorMeta:
    COMPILE_ERR = "compile error {}"
    UNKNOWN_MOD = "unknown Module {}"
    UNKNOWN_STMT = "unknown Stmt {}"
    UNKNOWN_EXPR = "unknown Expr {}"
    UNKNOWN_NUM = "unknown num {}"
    UNKNOWN_NAME_CONST = "unknown NameContext const {}"
    UNKNOWN_BINOP = "unknown BinOp {}"
    UNKNOWN_AUG_BINOP = "unknown Aug BinOp {}"
    UNKNOWN_UNARYOP = "unknown UnaryOp {}"
    UNKNOWN_LOOPTYPE = "unknown loop type {}"
    UNKNOWN_CMPOP = "unknown CompareOp {}"
    UNKNOWN_COMP = "unknown comprehension {}"
    INVALID_KCL_AST_MSG = "invalid KCL AST type"
    INVALID_PARAM_IN_ATTR = "invalid param {} in attribute expression"
    INVALID_PARAM_IN_SUBSCR = "invalid param {} in attribute subscript"
    INVALID_SUBSCRIPT_KIND = "invalid subscript kind {}"
    INVALID_TARGET_IN_LIST_TUPLE = (
        "invalid starred assignment target outside a list or tuple"
    )
    INVALID_STARRED_EXPR = "invalid starred expression outside assignment target "
    INVALID_TWO_STARRED_EXPR = "two starred expressions in assignment"
    INVALID_JMP_CALL = "opJmp called with non jump instruction"
    INVALID_NAME = "NameContext can't be None, True or False"
    INVALID_SYMBOL = "symbol can't be None, True or False"
    INVALID_SYMBOL_SCOPE = "invalid symbol scope {}"
    INVALID_NONE_OP = "opcode can't be none"
    INVALID_ARGED_OP_CODE = "opcode {} can't takes an argument"
    INVALID_OP_CODE = "opcode {} can't be emitted without an argument"
    INVALID_EXPR_CONTEXT = "invalid AST AssignStmt ExprContext value {}"
    INVALID_OP_POS = "invalid opcode pos {}"
    INVALID_MISSING_ARG = "missing arg in Arged opcode"
    INVALID_GLOBAL_IMPLICIT_SCOPE = "not expecting scopeGlobalImplicit in set qualname"
    INVALID_CLOSURE_FREE_SCOPE = (
        "invalid closure scope {} for free var {} in symbol table {}"
    )
    INVALID_GENERATORS = "invalid generators"
    INVALID_STRING_INTERPOLATION_ITEM = "invalid string interpolation item"
    INVALID_QUANTIFIER_OP = "invalid quantifier expression operation {}"
    SYMBOL_NOT_DEFINED = "name '{}' is not defined"
    UNEQUAL_OPS_AND_CMPS = "unequal ops and comparators in compare"
    UNEQUAL_DICT_KV_SIZE = "unequal dict keys len {} and values len {}"
    TOO_MANY_EXPRS_IN_STAR_UNPACK = "too many expressions in star-unpacking assignment"
    FAILED_SET_AUG_ASSIGN_CTX = "can't set context in AugAssign"
    NO_OPS_OR_CMPS = "no ops or comparators in compare"
    NO_SYMBOL_TABLE_FOR_AST = "no symbol table found for ast {}"
    DUPLICATED_KW = "duplicated keyword argument"


class SchemaConfigMeta:
    """
    SchemaConfigMeta defines the names of meta information
    """

    LINE = "$lineno"
    COLUMN = "$columnno"
    FILENAME = "$filename"



import sys
from typing import List, Dict, Tuple
import json
from kclvm.tools.lint.reporters.base_reporter import BaseReporter
from kclvm.tools.lint.message.message import Message


LEVEL_MAP = {"E": "error", "W": "waring", "C": "note"}


class SARIFMeta:
    VERSION = "2.1.0"
    SCHEMA = "https://docs.oasis-open.org/sarif/sarif/v2.1.0/cs01/schemas/sarif-schema-2.1.0.json"
    NAME = "kcl-lint"
    KCLLINT_VERSION = "0.0.1"
    INFORMATION_URI = "https://kusion-docs.com/docs/reference/cli/kcl/lint"


class Rule:
    def __init__(self, id: str, default: str, short: str):
        self.id = id
        self.messageStrings = {
            "default": {"text": default},
            "shortStrings": {"text": short},
        }


class Result:
    def __init__(self, m: Message):
        self.ruleId = m.msg_id
        self.level = LEVEL_MAP[m.msg_id[0]] if (m.msg_id[0] in LEVEL_MAP) else "note"
        self.message = {"id": "default", "arguments": m.arguments}
        self.locations = [
            {
                "physicalLocation": {
                    "artifactLocation": {"uri": m.file},
                    "region": {"startLine": m.pos[0], "startColumn": m.pos[1]},
                }
            }
        ]


class Tool:
    def __init__(self, ids: List[str], MSGS: Dict[str, Tuple[str, str]]):
        self.driver = {
            "name": SARIFMeta.NAME,
            "version": SARIFMeta.KCLLINT_VERSION,
            "informationUri": SARIFMeta.INFORMATION_URI,
            "rules": [Rule(id, MSGS[id][2], MSGS[id][1]) for id in ids],
        }


class SarifLog(object):
    """Static Analysis Results Format (SARIF) Version 2.1.0 JSON Schema: a standard format for the output of static analysis tools."""

    def __init__(self, msgs: List[Message], MSGS: Dict[str, Tuple[str, str]]):
        self.version = SARIFMeta.VERSION
        # self.$schema = SARIFMeta.SCHEMA
        msg_ids = list(set([m.msg_id for m in msgs]))
        self.runs = [
            {"tool": Tool(msg_ids, MSGS), "results": [Result(m) for m in msgs]}
        ]


class SARIFReporter(BaseReporter):
    def __init__(self, linter, output=None, encoding=None):
        self.name = "sarif_reporter"
        self.output_file = linter.config.output_path
        super().__init__(linter, output, encoding)

    def print_msg(self, msgs: List[Message] = None):
        assert self.output_file
        sarif_log = SarifLog(msgs, self.linter.MSGS)
        json_str = json.dumps(
            sarif_log, default=lambda o: o.__dict__, sort_keys=True, indent=2
        )
        with open(self.output_file, "w") as f:
            current = sys.stdout
            sys.stdout = f
            print(json_str)
            sys.stdout = current



# Copyright 2021 The KCL Authors. All rights reserved.

import typing
import ast
import os

import kclvm.kcl.error as kcl_error
import kclvm.internal.util as util
import kclvm.config


class KclOptionElem:
    def __init__(self, name: str):
        self.name = name
        self.defined = False
        self.value_type = ""  # bool/int/float/str/list/dict
        self.required = False
        self.default = None
        self.help = ""
        self.file = ""
        self.line = 0

        self.inited = False
        self.value = None

    def get_help(self, verbose_mode=0) -> str:
        name = self.name
        default = self.default if self.default else "?"

        type_and_required = ""
        if self.value_type != "" and self.required:
            type_and_required = f" ({self.value_type},required)"
        elif self.value_type != "":
            type_and_required = f" ({self.value_type})"
        elif self.required:
            type_and_required = " (required)"

        if verbose_mode > 1 and self.file and self.line > 0:
            filename = os.path.relpath(self.file, os.getcwd())
            return f"{name}={default}{type_and_required} {self.help} ({filename}:{self.line})".strip()
        else:
            return f"{name}={default}{type_and_required} {self.help}".strip()

    def __str__(self) -> str:
        return self.get_help()


class _KclOptionDict:
    def __init__(self):
        self.m: typing.Dict[str, KclOptionElem] = {}  # map[name]KclOptionElem
        self.reset()

    def reset(self):
        self.m = {}  # map[name]KclOptionElem

    def len(self) -> int:
        return len(self.m)

    def get_dict(self):
        return self.m  # map[name]KclOptionElem

    def keys(self) -> list:
        return list(self.m.keys())

    def has_key(self, name: str) -> bool:
        return name in self.m

    @classmethod
    def _check_value_type(cls, value_type: str, value: typing.Any) -> bool:
        if not value_type or value is None:
            return True
        return cls._get_typed_value(value_type, value) is not None

    # noinspection PyBroadException
    @classmethod
    def _get_typed_value(cls, value_type: str, value) -> typing.Any:
        if not value_type:
            return value
        if value is None:
            return None

        if value_type == "bool":
            return True if str(value).lower() in ("yes", "true", "1") else bool(value)
        if value_type == "int":
            try:
                return int(value)
            except Exception:
                return None
        if value_type == "float":
            try:
                return float(value)
            except Exception:
                return None
        if value_type == "str":
            return str(value)
        if value_type == "list":
            if isinstance(value, list):
                return value
            try:
                result = ast.literal_eval(value)
                return result if type(result) is list else None
            except Exception:
                return None
        if value_type == "dict":
            if isinstance(value, dict):
                return value
            try:
                result = ast.literal_eval(value)
                return result if type(result) is dict else None
            except Exception:
                return None

        # unknown type
        kcl_error.report_exception(
            err_type=kcl_error.ErrType.EvaluationError_TYPE,
            arg_msg=f"unknown type: {value_type}",
        )

    def option(
        self,
        name: str,
        *,
        value_type="",
        required=False,
        default=None,
        help="",
        file="",
        line=0,
    ) -> typing.Any:
        if value_type and value_type not in [
            "bool",
            "int",
            "float",
            "str",
            "list",
            "dict",
        ]:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.EvaluationError_TYPE,
                arg_msg=f"'{value_type}' must one of bool/int/float/str/list/dict or empty string",
            )

        opt = self.get(name)
        opt.defined = True

        if value_type:
            opt.value_type = value_type
        if required:
            opt.required = required
        if default is not None:
            opt.default = default
        if help:
            opt.help = help
        if file:
            opt.file = file
        if line > 0:
            opt.line = line

        if opt.value is None:
            opt.value = opt.default

        if value_type and opt.value is not None:
            raw_value = opt.value
            opt.value = self._get_typed_value(value_type, opt.value)
            if opt.value is None:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    arg_msg=f"cannot use '{raw_value}' as type '{value_type}'",
                )

        self.m[name] = opt
        kcl_option_exec()  # FIXME: filename, lineno, colno info
        return opt.value  # Do type conversion?

    def init_value(self, *d_list, name: str = "", value=None, **attrs):
        def _init_kv(name_: str, value_: str):
            opt = self.get(name_)
            opt.value = value_
            opt.inited = True
            self.m[name_] = opt

        for d in d_list:
            kv = d.split("=")
            if len(kv) != 2 or not kv[0]:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    arg_msg=f"invalid args: {d}",
                )
            _init_kv(kv[0], kv[1])

        for name, value in attrs.items():
            _init_kv(name, value)

        if name:
            _init_kv(name, value)

    def get(self, name: str) -> KclOptionElem:
        if name in self.m:
            return self.m[name]
        else:
            return KclOptionElem(name)

    def check(self, *, check_required=True, check_none=False, verbose_mode=0):
        for name, v in self.m.items():
            if verbose_mode > 1 and not v.defined:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    file_msgs=[kcl_error.ErrFileMsg(line_no=1)],
                    arg_msg=f"invalid '-D {name}={v.value}', option('{name}') undefined",
                )
            if check_required and v.required and (not v.inited):
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    file_msgs=[kcl_error.ErrFileMsg(filename=v.file, line_no=v.line)],
                    arg_msg=f"option('{name}') must be initialized, try '-D {name}=?' argument",
                )
            if check_none and not v.value:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    file_msgs=[kcl_error.ErrFileMsg(filename=v.file, line_no=v.line)],
                    arg_msg=f"option('{name}') is None",
                )
            if not self._check_value_type(v.value_type, v.value):
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.EvaluationError_TYPE,
                    file_msgs=[kcl_error.ErrFileMsg(line_no=1)],
                    arg_msg=f"cannot use '-D {name}={v.value}' as type {v.value_type}",
                )

        return None

    def help(self, *, prefix: str = "", verbose_mode=0) -> str:
        if len(self.m) == 0:
            return ""

        msg = "option list:\n"
        for name in self.keys():
            msg += f"{prefix}{self.get(name).get_help(verbose_mode)}\n"
        return msg[:-1]  # remove \n


_kcl_option_dict = _KclOptionDict()


def kcl_option_dict():
    return _kcl_option_dict.get_dict()


def kcl_option_help(verbose_mode=0) -> str:
    return _kcl_option_dict.help(prefix="  -D ", verbose_mode=verbose_mode)


def kcl_option_check(verbose_mode=0):
    _kcl_option_dict.check(verbose_mode=verbose_mode)


def kcl_option_init(*args, name: str = "", value=None, **attrs):
    _kcl_option_dict.init_value(*args, name=name, value=value, **attrs)


def kcl_option(
    name: str, *, type="", required=False, default=None, help="", file="", line=0
) -> typing.Any:
    return _kcl_option_dict.option(
        name,
        value_type=type,
        required=required,
        default=default,
        help=help,
        file=file,
        line=line,
    )


def kcl_option_reset():
    _kcl_option_dict.reset()


def kcl_option_exec():
    if kclvm.config.list_option_mode > 0:
        kclvm.config.options_help_message = kcl_option_help(
            verbose_mode=kclvm.config.list_option_mode
        )
    else:
        kcl_option_check()


def kcl_option_init_all():
    if kclvm.config.arguments:
        for _k, _v in util.merge_option_same_keys(kclvm.config.arguments).items():
            kcl_option_init(name=_k, value=_v)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: modfile/modfile.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15modfile/modfile.proto\x12\rkclvm.modfile"\xa2\x01\n\nKclModFile\x12\x0c\n\x04root\x18\x01 \x01(\t\x12\x10\n\x08root_pkg\x18\x02 \x01(\t\x12\x36\n\x05\x62uild\x18\x03 \x01(\x0b\x32\'.kclvm.modfile.KclModFile_build_section\x12<\n\x08\x65xpected\x18\x04 \x01(\x0b\x32*.kclvm.modfile.KclModFile_expected_section"_\n\x18KclModFile_build_section\x12\x18\n\x10\x65nable_pkg_cache\x18\x01 \x01(\x08\x12\x19\n\x11\x63\x61\x63hed_pkg_prefix\x18\x02 \x01(\t\x12\x0e\n\x06target\x18\x03 \x01(\t"\x98\x01\n\x1bKclModFile_expected_section\x12\x16\n\x0emin_build_time\x18\x01 \x01(\t\x12\x16\n\x0emax_build_time\x18\x02 \x01(\t\x12\x15\n\rkclvm_version\x18\x03 \x01(\t\x12\x1a\n\x12kcl_plugin_version\x18\x04 \x01(\t\x12\x16\n\x0eglobal_version\x18\x05 \x01(\tb\x06proto3'
)


_KCLMODFILE = DESCRIPTOR.message_types_by_name["KclModFile"]
_KCLMODFILE_BUILD_SECTION = DESCRIPTOR.message_types_by_name["KclModFile_build_section"]
_KCLMODFILE_EXPECTED_SECTION = DESCRIPTOR.message_types_by_name[
    "KclModFile_expected_section"
]
KclModFile = _reflection.GeneratedProtocolMessageType(
    "KclModFile",
    (_message.Message,),
    {
        "DESCRIPTOR": _KCLMODFILE,
        "__module__": "modfile.modfile_pb2"
        # @@protoc_insertion_point(class_scope:kclvm.modfile.KclModFile)
    },
)
_sym_db.RegisterMessage(KclModFile)

KclModFile_build_section = _reflection.GeneratedProtocolMessageType(
    "KclModFile_build_section",
    (_message.Message,),
    {
        "DESCRIPTOR": _KCLMODFILE_BUILD_SECTION,
        "__module__": "modfile.modfile_pb2"
        # @@protoc_insertion_point(class_scope:kclvm.modfile.KclModFile_build_section)
    },
)
_sym_db.RegisterMessage(KclModFile_build_section)

KclModFile_expected_section = _reflection.GeneratedProtocolMessageType(
    "KclModFile_expected_section",
    (_message.Message,),
    {
        "DESCRIPTOR": _KCLMODFILE_EXPECTED_SECTION,
        "__module__": "modfile.modfile_pb2"
        # @@protoc_insertion_point(class_scope:kclvm.modfile.KclModFile_expected_section)
    },
)
_sym_db.RegisterMessage(KclModFile_expected_section)

if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _KCLMODFILE._serialized_start = 41
    _KCLMODFILE._serialized_end = 203
    _KCLMODFILE_BUILD_SECTION._serialized_start = 205
    _KCLMODFILE_BUILD_SECTION._serialized_end = 300
    _KCLMODFILE_EXPECTED_SECTION._serialized_start = 303
    _KCLMODFILE_EXPECTED_SECTION._serialized_end = 455
# @@protoc_insertion_point(module_scope)



from typing import Any

from kclvm.api.object import KCLNumberMultiplierObject
from kclvm.api.object.internal import Undefined, UndefinedType

import kclvm.kcl.info as kcl_info


def filter_fields(
    value: Any, ignore_private: bool = False, ignore_none: bool = False
) -> Any:
    """Remove private attributes start with '_' and None value in data"""
    if not value:
        return value
    if value is Undefined or isinstance(value, UndefinedType):
        return value
    if isinstance(value, KCLNumberMultiplierObject):
        return str(value)
    if isinstance(value, list):
        return [
            filter_fields(_v, ignore_private, ignore_none)
            for _v in value
            if ignore_none and _v is not None or not ignore_none
            if not isinstance(_v, UndefinedType)
        ]
    elif isinstance(value, dict):
        return {
            filter_fields(_k, ignore_private, ignore_none): filter_fields(
                _v, ignore_private, ignore_none
            )
            for _k, _v in value.items()
            if not kcl_info.isprivate_field(_k) and ignore_private or not ignore_private
            if ignore_none and _v is not None or not ignore_none
            if not isinstance(_v, UndefinedType)
        }
    elif isinstance(value, (int, float, str, bool)):
        return value
    else:
        raise Exception("Invalid KCL Object")



"""
https://microsoft.github.io/language-server-protocol/specification#textDocument_documentSymbol
DocumentSymbol[] which is a hierarchy of symbols found in a given text document.
class DocumentSymbol(Model):
    name: str 	                                       # The name of this symbol. Will be displayed in the user interface
                                                         and therefore must not be an empty string or a string only
                                                         consisting of white spaces.
    kind: SymbolKind                                   # The kind of this symbol.
    range: Range                                       # The range enclosing this symbol not including leading/trailing
                                                         whitespace but everything else like comments. This information
                                                         is typically used to determine if the clients cursor is inside
                                                         the symbol to reveal in the symbol in the UI.
    selection_range: Range                             # The range that should be selected and revealed when this symbol
                                                         is being picked, e.g. the name of a function. Must be contained
                                                         by the `range`.
    detail: Optional[str] = None                       # More detail for this symbol, e.g the signature of a function.
    children: Optional[List['DocumentSymbol']] = None  # Children of this symbol, e.g. properties of a class.
    deprecated: Optional[bool] = False                 # Indicates if this symbol is deprecated.

In KCl, we select variables and schema definitions as symbols，, and the schema attributes and mixins in the schema will
be the child nodes of the schema, e.g:

a = b = 1                        a
schema Person:                   b
    mixin [                      Person
        nameMixin                  mixin
    ]                       ->       nameMixin
    age: int = 1                   age
                                 person
person = Person{
    ...
}
"""
from typing import List
from pygls.lsp.types.language_features.document_symbol import DocumentSymbol, SymbolKind
from pygls.lsp.types.basic_structures import Range, Position

import kclvm.kcl.ast as ast
from kclvm.tools.langserver.common import file_to_ast


def range_check(s: DocumentSymbol):
    """
    DocumentSymbol.selection_range must be contained by the DocumentSymbol.range
        a = { ... }
        ^^        ^
        ││        └-range_end
        │└---selection_range_end
        │
        range_start, selection_range_start

    """
    assert isinstance(s, DocumentSymbol)
    range = s.range
    selection_range = s.selection_range
    assert (
        (range.start.line <= selection_range.start.line)
        & (selection_range.start.line <= selection_range.end.line)
        & (selection_range.end.line <= range.end.line)
    )
    if range.start.line == selection_range.start.line:
        assert selection_range.start.character >= range.start.character
    if selection_range.end.line == range.end.line:
        assert selection_range.start.character >= range.start.character
    if s.children:
        for child in s.children:
            range_check(child)


def ast_position_to_range(node: ast.AST) -> Range:
    assert isinstance(node, ast.AST)
    return Range(
        start=Position(line=node.line - 1, character=node.column - 1),
        end=Position(line=node.end_line - 1, character=node.end_column - 1),
    )


def identifier_to_document_symbol(node: ast.Identifier) -> DocumentSymbol:
    assert isinstance(node, ast.Identifier)
    range = Range(
        start=Position(line=node.line - 1, character=node.column - 1),
        end=Position(line=node.end_line - 1, character=node.end_column - 1),
    )
    return DocumentSymbol(
        name=".".join(node.names),
        range=range,
        selection_range=range,
        kind=SymbolKind.Variable,
    )


def schema_attr_to_document_symbol(node: ast.SchemaAttr) -> DocumentSymbol:
    assert isinstance(node, ast.SchemaAttr)
    return DocumentSymbol(
        name=node.name,
        range=ast_position_to_range(node),
        kind=SymbolKind.Property,
        selection_range=ast_position_to_range(node.name_node),
    )


def schema_stmt_to_document_symbol(node: ast.SchemaStmt) -> DocumentSymbol:
    assert isinstance(node, ast.SchemaStmt)
    symbol = DocumentSymbol(
        name=node.name,
        kind=SymbolKind.Struct,
        range=ast_position_to_range(node),
        selection_range=ast_position_to_range(node.name_node),
        children=[],
    )
    if len(node.mixins):
        range = Range(
            start=Position(line=node.line, character=4),
            end=Position(line=node.line, character=9),
        )
        symbol.children.append(
            DocumentSymbol(
                name="mixin",
                kind=SymbolKind.Property,
                range=range,
                selection_range=range,
                children=[identifier_to_document_symbol(id) for id in node.mixins],
            )
        )
    symbol.children += [
        schema_attr_to_document_symbol(attr) for attr in node.GetAttrList()
    ]
    return symbol


def assign_stmt_to_document_symbol(node: ast.AssignStmt) -> DocumentSymbol:
    assert isinstance(node, ast.AssignStmt)
    id_symbols = []
    for identifier in node.targets:
        symbol = identifier_to_document_symbol(identifier)
        symbol.range = ast_position_to_range(node)
        id_symbols.append(symbol)
    return id_symbols


def document_symbol(file: str, code: str = None) -> List[DocumentSymbol]:
    symbols = []
    module = file_to_ast(file, code)
    if not module or not isinstance(module, ast.Module):
        return []
    for stmt in module.body or []:
        if isinstance(stmt, ast.SchemaStmt):
            symbols.append(schema_stmt_to_document_symbol(stmt))
        if isinstance(stmt, ast.AssignStmt):
            symbols.extend(assign_stmt_to_document_symbol(stmt))
    for s in symbols:
        range_check(s)
    return symbols



from enum import IntEnum
from typing import Union

import kclvm.kcl.ast.ast as ast


class OpPrecedence(IntEnum):
    """Operator precedence in grammar

    assign_stmt:  // precedence 0
    or_test: and_test (L_OR and_test)*  // precedence 1
    and_test: not_test (L_AND not_test)*  // precedence 2
    not_test: L_NOT not_test | comparison  // precedence 3
    comparison: expr (comp_op expr)*  // precedence 4
    expr: xor_expr (OR xor_expr)*  // precedence 5
    xor_expr: and_expr (XOR and_expr)*  // precedence 6
    and_expr: shift_expr (AND shift_expr)*  // precedence 7
    shift_expr: arith_expr ((SHIFT_LEFT|SHIFT_RIGHT) arith_expr)*  // precedence 8
    arith_expr: term ((PLUS|MINUS) term)*  // precedence 9
    term: factor ((MULTIPLY|DIVIDE|MOD|DOUBLE_DIVIDE) factor)*  // precedence 10
    factor: (PLUS|MINUS|NOT) factor | power  // precedence 11
    power: primary_expr (DOUBLE_STAR factor)?  // precedence 12
    """

    LOWEST = 0
    HIGHEST = 12


OP_PREC_MAP = {
    ast.AugOp.Assign: 0,
    ast.AugOp.Add: 0,
    ast.AugOp.Sub: 0,
    ast.AugOp.Mul: 0,
    ast.AugOp.Div: 0,
    ast.AugOp.Mod: 0,
    ast.AugOp.Pow: 0,
    ast.AugOp.LShift: 0,
    ast.AugOp.RShift: 0,
    ast.AugOp.BitOr: 0,
    ast.AugOp.BitXor: 0,
    ast.AugOp.BitAnd: 0,
    ast.AugOp.FloorDiv: 0,
    ast.BinOp.Add: 9,
    ast.BinOp.Sub: 9,
    ast.BinOp.Mul: 10,
    ast.BinOp.Div: 10,
    ast.BinOp.Mod: 10,
    ast.BinOp.Pow: 12,
    ast.BinOp.LShift: 8,
    ast.BinOp.RShift: 8,
    ast.BinOp.BitOr: 6,
    ast.BinOp.BitXor: 5,
    ast.BinOp.BitAnd: 7,
    ast.BinOp.FloorDiv: 10,
    ast.BinOp.As: 4,
    ast.BinOp.And: 2,
    ast.BinOp.Or: 1,
    ast.CmpOp.Eq: 4,
    ast.CmpOp.NotEq: 4,
    ast.CmpOp.Lt: 4,
    ast.CmpOp.LtE: 4,
    ast.CmpOp.Gt: 4,
    ast.CmpOp.GtE: 4,
    ast.CmpOp.Is: 4,
    ast.CmpOp.In: 4,
    ast.CmpOp.Not: 4,
    ast.CmpOp.IsNot: 4,
    ast.CmpOp.NotIn: 4,
    ast.UnaryOp.UAdd: 11,
    ast.UnaryOp.USub: 11,
    ast.UnaryOp.Invert: 11,
    ast.UnaryOp.Not: 3,
}


def precedence(op: Union[ast.BinOp, ast.AugOp, ast.UnaryOp, ast.CmpOp]) -> int:
    """Return KCL operator precedence"""
    if not op:
        return int(OpPrecedence.LOWEST)

    # x ** y -> 12
    if op == ast.BinOp.Pow:
        return int(OpPrecedence.HIGHEST)

    # +x, -x, ~x -> 11
    if op == ast.UnaryOp.UAdd or op == ast.UnaryOp.USub or op == ast.UnaryOp.Invert:
        return OpPrecedence.HIGHEST - 1

    # x * y, x / y, x // y, x % y -> 10
    if (
        op == ast.BinOp.Mul
        or op == ast.BinOp.Div
        or op == ast.BinOp.FloorDiv
        or op == ast.BinOp.Mod
    ):
        return OpPrecedence.HIGHEST - 2

    # x + y, x - y -> 9
    if op == ast.BinOp.Add or op == ast.BinOp.Sub:
        return OpPrecedence.HIGHEST - 3

    # x >> y, x << y -> 8
    if op == ast.BinOp.LShift or op == ast.BinOp.RShift:
        return OpPrecedence.HIGHEST - 4

    # x & y -> 7
    if op == ast.BinOp.BitAnd:
        return OpPrecedence.HIGHEST - 5

    # x | y -> 6
    if op == ast.BinOp.BitOr:
        return OpPrecedence.HIGHEST - 6

    # x ^ y -> 5
    if op == ast.BinOp.BitXor:
        return OpPrecedence.HIGHEST - 7

    # x > y, x < y, etc. -> 4
    if isinstance(op, ast.CmpOp):
        return OpPrecedence.HIGHEST - 8

    # not x -> 3
    if op == ast.UnaryOp.Not:
        return OpPrecedence.HIGHEST - 9

    # x and y -> 2
    if op == ast.BinOp.And:
        return OpPrecedence.HIGHEST - 10

    # x or y -> 1
    if op == ast.BinOp.Or:
        return OpPrecedence.HIGHEST - 11

    # x = y, x += y, x -= y -> 0
    if isinstance(op, ast.AugOp):
        assert OpPrecedence.HIGHEST - 12 == OpPrecedence.LOWEST
        return OpPrecedence.HIGHEST - 12

    return int(OpPrecedence.LOWEST)



"""The `protobuf` module defines functionality for parsing protocol buffer
definitions and instances.

Proto definition mapping follows the guidelines of mapping Proto to JSON as
discussed in https://developers.google.com/protocol-buffers/docs/proto3, and
carries some of the mapping further when possible with KCL.

The following type mappings of definitions apply:

    Proto type        KCL type/def     Comments
    message           schema           Message fields become KCL attribute, whereby
                                       names are mapped to lowerCamelCase.
    enum              e1 | e2 | ...    Where ex are strings. A separate mapping is
                                       generated to obtain the numeric values.
    oneof             e1 | e2 | ...    KCL union type.
    map<K, V>         {K:V}
    repeated V        [V]
    bool              bool
    string            str
    bytes             str              A base64-encoded string when converted to JSON.
    int32, fixed32    int              An integer with bounds as defined by int.
    uint32            int              An integer with bounds as defined by int.
    int64, fixed64    int              An integer with bounds as defined by int.
    uint64            int              An integer with bounds as defined by int.
    float             float            A number with bounds as defined by float.
    double            float            A number with bounds as defined by float.
    Struct            schema
    Value             any
    google.proto.Any  any
    ListValue         [...]
    NullValue         any
    BoolValue         bool
    StringValue       str
    NumberValue       float
    StringValue       str
    Empty             any
    Timestamp         float
    Duration          int

Protobuf definitions can be annotated with KCL constraints that are included
in the generated KCL:
    (kcl.opt)     FieldOptions
        val          str           KCL attribute default value.
        optional     bool          Whether the KCL attribute is optional
        final        bool          Whether the KCL attribute is final

:copyright: Copyright 2020 The KCL Authors. All rights reserved.
"""

from .parser import (
    Import,
    Package,
    Option,
    Field,
    OneOfField,
    OneOf,
    Map,
    Reserved,
    Range,
    EnumField,
    Enum,
    Message,
    Service,
    Rpc,
    Proto,
    ImportOption,
    Type,
    KeyType,
    parse_code,
)
from .protobuf import (
    protobuf_to_kcl,
)

__all__ = [
    "Import",
    "Package",
    "Option",
    "Field",
    "OneOfField",
    "OneOf",
    "Map",
    "Reserved",
    "Range",
    "EnumField",
    "Enum",
    "Message",
    "Service",
    "Rpc",
    "Proto",
    "ImportOption",
    "Type",
    "KeyType",
    "parse_code",
    "protobuf_to_kcl",
]



# Copyright 2021 The KCL Authors. All rights reserved.

import io
import re
import sys
from enum import IntEnum
from dataclasses import dataclass
from typing import List, Dict, Union, Tuple, Optional

import kclvm.kcl.ast as ast
import kclvm.compiler.astutil.fix as fix

# ---------------------------------------------------
# Constants
# ---------------------------------------------------


class Indentation(IntEnum):
    Indent = 0
    Dedent = 1
    Newline = 2
    IndentWithNewline = 3
    DedentWithNewline = 4
    Fill = 5


_INVALID_AST_MSG = "Invalid AST Node"
_TEMP_ROOT = "<root>"

WHITESPACE = " "
TAB = "\t"
NEWLINE = "\n"
COMMA_WHITESPACE = ast.TokenValue.COMMA + WHITESPACE
IDENTIFIER_REGEX = r"^\$?[a-zA-Z_]\w*$"
QUANT_OP_TOKEN_VAL_MAPPING = {
    ast.QuantOperation.ALL: ast.TokenValue.ALL,
    ast.QuantOperation.ANY: ast.TokenValue.ANY,
    ast.QuantOperation.MAP: ast.TokenValue.MAP,
    ast.QuantOperation.FILTER: ast.TokenValue.FILTER,
}

# ---------------------------------------------------
# Printer config
# ---------------------------------------------------


@dataclass
class Config:
    tab_len: int = 4
    indent_len: int = 4
    use_spaces: bool = True
    is_fix: bool = False


# ---------------------------------------------------
# Printer
# ---------------------------------------------------


class BasePrinter(ast.TreeWalker):
    def __init__(self, config: Config, out: io.TextIOBase):
        self.output: str = ""
        self.indent = 0
        self.cfg: Config = config
        self.out: io.TextIOBase = out
        # Print comments
        self.last_ast_line: int = 0
        self.comments: List[ast.Comment] = []
        self.import_spec: Dict[str, str] = {}

    # Base walker functions

    def get_node_name(self, t: ast.AST):
        """Get the ast.AST node name"""
        assert isinstance(t, ast.AST)
        return t.type

    def generic_walk(self, t: ast.AST):
        """Called if no explicit walker function exists for a node."""
        if not isinstance(t, ast.AST):
            raise Exception(_INVALID_AST_MSG, t)
        else:
            self.walk(t)

    def write_ast_comments(self, t: ast.AST):
        if not t or not isinstance(t, ast.AST):
            return
        if t.line and t.line > self.last_ast_line:
            self.last_ast_line = t.line
            if self.comments:
                index = -1
                for i, comment in enumerate(self.comments):
                    if comment.line <= t.line:
                        index = i
                        self.write(comment.text + NEWLINE)
                        self.fill()
                    else:
                        break
                if index >= 0:
                    del self.comments[: index + 1]

    # -----------------
    # Expr and Stmt walker functions
    # -----------------

    def expr(self, expr: ast.Expr):
        if not expr:
            return
        self.write_ast_comments(expr)
        self.generic_walk(expr)

    def stmt(self, stmt: ast.Stmt):
        if not stmt:
            return
        self.fill()
        self.write_ast_comments(stmt)
        self.generic_walk(stmt)

    def stmts(self, stmts: List[ast.Stmt]):
        for stmt in stmts or []:
            self.stmt(stmt)

    def exprs(self, exprs: List[ast.Expr]):
        for expr in exprs or []:
            self.expr(expr)

    # --------------------------
    # Indent and scope functions
    # --------------------------

    def enter(self):
        self.indent += 1

    def leave(self):
        self.indent -= 1

    # --------------------------
    # Write functions
    # --------------------------

    @staticmethod
    def interleave(inter, f, seq):
        """Call f on each item in seq, calling inter() in between."""
        if not seq:
            return
        seq = iter(seq)
        try:
            f(next(seq))
        except StopIteration:
            pass
        else:
            for x in seq:
                inter()
                f(x)

    def write(self, text: str = ""):
        self.out.write(text)

    def writeln(self, text: str = ""):
        self.write(text + NEWLINE)
        self.fill()

    def write_with_spaces(self, text: str = ""):
        self.write(" " + text + " ")

    def fill(self, text: str = ""):
        """Indent a piece of text, according to the current indentation level"""
        if self.cfg.use_spaces:
            self.write(WHITESPACE * self.indent * self.cfg.indent_len + text)
        else:
            self.write(TAB * self.indent + text)

    def print(self, *values):
        for value in values or []:
            if isinstance(value, ast.Expr):
                self.expr(value)
            elif isinstance(value, ast.Stmt):
                self.stmt(value)
            elif isinstance(value, ast.AST):
                self.generic_walk(value)
            elif value == Indentation.Indent:
                self.enter()
            elif value == Indentation.Dedent:
                self.leave()
            elif value == Indentation.Newline:
                self.writeln()
            elif value == Indentation.IndentWithNewline:
                self.enter()
                self.writeln()
            elif value == Indentation.DedentWithNewline:
                self.leave()
                self.writeln()
            elif value == Indentation.Fill:
                self.fill()
            elif isinstance(value, str):
                self.write(value)
            elif isinstance(value, (int, float, bool)):
                self.write(str(value))

    def print_ast(self, t: ast.AST):
        if not t:
            return
        if isinstance(t, ast.Module) and self.cfg.is_fix:
            fix.fix_and_get_module_import_list(
                _TEMP_ROOT, t, is_fix=True, reversed=True
            )
        self.walk(t)
        for comment in self.comments or []:
            self.write(comment.text + NEWLINE)
            self.fill()


class Printer(BasePrinter):
    def __init__(self, config: Config, out: io.TextIOBase):
        super().__init__(config, out)

    def walk_Module(self, t: ast.Module):
        """ast.AST: Module

        Parameters
        ----------
        - body: List[Stmt]
        """
        assert isinstance(t, ast.Module)
        self.comments = t.comments
        self.stmts(t.body)

    def walk_ExprStmt(self, t: ast.ExprStmt):
        """ast.AST: ExprStmt

        Parameters
        ----------
        - exprs: List[Expr]
        """
        assert isinstance(t, ast.ExprStmt)
        self.interleave(lambda: self.write(COMMA_WHITESPACE), self.expr, t.exprs)
        self.writeln()

    def walk_AssertStmt(self, t: ast.AssertStmt):
        """ast.AST: AssertStmt

        Parameters
        ----------
        - test: Expr
        - if_cond: Expr
        - msg: Expr
        """
        assert isinstance(t, ast.AssertStmt) and t.test
        self.print(
            ast.TokenValue.ASSERT,
            WHITESPACE,
            t.test,
        )
        if t.if_cond:
            self.print(WHITESPACE, ast.TokenValue.IF, WHITESPACE, t.if_cond)
        if t.msg:
            self.print(
                COMMA_WHITESPACE,
                t.msg,
            )
        self.print(NEWLINE)

    def walk_IfStmt(self, t: ast.IfStmt):
        """ast.AST: IfStmt

        Parameters
        ----------
        - cond: Expr
        - body: List[Stmt]
        - elif_cond: List[Expr]
        - elif_body: List[List[Stmt]]
        - else_body: List[Stmt]
        """
        assert isinstance(t, ast.IfStmt)
        assert t.cond
        assert t.body
        self.print(
            ast.TokenValue.IF,
            WHITESPACE,
            t.cond,
            ast.TokenValue.COLON,
            NEWLINE,
            Indentation.Indent,
        )
        self.stmts(t.body)
        self.print(Indentation.Dedent)
        if t.elif_cond:
            for cond, body in zip(t.elif_cond, t.elif_body):
                # Nested if statements need to be considered,
                # so `elif` needs to be preceded by the current indentation.
                self.print(
                    Indentation.Fill,
                    ast.TokenValue.ELIF,
                    WHITESPACE,
                    cond,
                    ast.TokenValue.COLON,
                    NEWLINE,
                    Indentation.Indent,
                )
                self.stmts(body)
                self.print(Indentation.Dedent)
        if t.else_body:
            # Nested if statements need to be considered,
            # so `else` needs to be preceded by the current indentation.
            self.print(
                Indentation.Fill,
                ast.TokenValue.ELSE,
                ast.TokenValue.COLON,
                NEWLINE,
                Indentation.Indent,
            )
            self.stmts(t.else_body)
            self.print(Indentation.Dedent)

    def walk_ImportStmt(self, t: ast.ImportStmt):
        """ast.AST: ImportStmt

        Parameters
        ---------
        - path: str
        - name: str
        - asname: str
        """
        assert isinstance(t, ast.ImportStmt)
        assert t.pkg_name
        self.print(
            ast.TokenValue.IMPORT,
            WHITESPACE,
            t.path,
        )
        if t.asname:
            self.print(
                WHITESPACE,
                ast.TokenValue.AS,
                WHITESPACE,
                t.asname,
            )
        self.import_spec[t.path] = t.pkg_name
        self.writeln()

    def walk_SchemaStmt(self, t: ast.SchemaStmt):
        """ast.AST: SchemaStmt

        Parameters
        ----------
        - doc: str
        - name: str
        - parent_name: Identifier
        - is_mixin: bool
        - is_protocol: bool
        - args: Arguments
        - settings: dict
        - mixins: List[str]
        - body: List[Union[SchemaAttr, Stmt]]
        - decorators: List[Decorator]
        - checks: List[CheckExpr]
        - for_host_name: Optional[Identifier] = None
        - index_signature: Optional[SchemaIndexSignature] = None
        """
        assert isinstance(t, ast.SchemaStmt)
        self.exprs(t.decorators)
        tok = ast.TokenValue.SCHEMA
        if t.is_mixin:
            tok = ast.TokenValue.MIXIN
        elif t.is_protocol:
            tok = ast.TokenValue.PROTOCOL
        self.print(
            tok,
            WHITESPACE,
        )
        self.print(t.name)
        if t.args:
            self.print(
                ast.TokenValue.LEFT_BRACKETS,
                t.args,
                ast.TokenValue.RIGHT_BRACKETS,
            )
        if t.parent_name:
            self.print(
                ast.TokenValue.LEFT_PARENTHESES,
                t.parent_name,
                ast.TokenValue.RIGHT_PARENTHESES,
            )
        if t.for_host_name:
            self.print(
                WHITESPACE,
                ast.TokenValue.FOR,
                WHITESPACE,
                t.for_host_name,
            )
        self.print(
            ast.TokenValue.COLON,
            NEWLINE,
            Indentation.Indent,  # Schema Stmt indent
        )
        if t.doc:
            self.print(
                Indentation.Fill,
                '"""',
                t.doc.replace('"', '\\"'),
                '"""',
                NEWLINE,
            )
        if t.mixins:
            self.print(
                Indentation.Fill,
                ast.TokenValue.MIXIN,
                WHITESPACE,
                ast.TokenValue.LEFT_BRACKETS,
                Indentation.IndentWithNewline,
            )
            self.interleave(
                lambda: self.print(ast.TokenValue.COMMA, Indentation.Newline),
                self.print,
                t.mixins,
            )
            self.print(
                Indentation.Dedent,
                Indentation.Newline,
                ast.TokenValue.RIGHT_BRACKETS,
                NEWLINE,
            )
        if t.index_signature:
            self.print(t.index_signature)
            self.print(NEWLINE)
        self.stmts(t.body)
        self.write(NEWLINE)
        if t.checks:
            self.print(
                Indentation.Fill,
                ast.TokenValue.CHECK,
                ast.TokenValue.COLON,
                Indentation.IndentWithNewline,  # Schema check indent
            )
            self.interleave(
                lambda: self.print(Indentation.Newline),
                self.print,
                t.checks,
            )
            self.write(NEWLINE)
            # Schema check dedent
            self.print(Indentation.Dedent)
            self.write(NEWLINE)
        # Schema Stmt dedent
        self.print(Indentation.Dedent)

    def walk_SchemaIndexSignature(self, t: ast.SchemaIndexSignature):
        """ast.AST: SchemaIndexSignature

        Parameters
        ----------
        - key_name: Optional[str]
        - key_type: Optional[str]
        - value_type: Optional[str]
        - value: Optional[Expr]
        - any_other: bool
        - name_node: Optional[Name]
        - value_type_node: Optional[Type]
        """
        self.write(ast.TokenValue.LEFT_BRACKETS)
        if t.any_other:
            self.write("...")
        if t.key_name:
            self.print(
                t.key_name,
                ast.TokenValue.COLON,
                WHITESPACE,
            )
        self.print(
            t.key_type,
            ast.TokenValue.RIGHT_BRACKETS,
            ast.TokenValue.COLON,
            WHITESPACE,
            t.value_type,
        )
        if t.value:
            self.print(WHITESPACE, ast.TokenValue.ASSIGN, WHITESPACE, t.value)

    def walk_RuleStmt(self, t: ast.RuleStmt):
        """ast.AST: RuleStmt

        Parameters
        ----------
        self.doc: str = ""
        self.name: str = ""
        self.parent_rules: List[Identifier] = []
        self.decorators: List[Decorator] = []
        self.checks: List[CheckExpr] = []
        self.name_node: Optional[Name] = None
        self.args: Optional[Arguments] = None
        self.for_host_name: Optional[Identifier] = None
        """
        assert isinstance(t, ast.RuleStmt)
        self.exprs(t.decorators)
        self.print(
            ast.TokenValue.RULE,
            WHITESPACE,
        )
        self.print(t.name)
        if t.args:
            self.print(
                ast.TokenValue.LEFT_BRACKETS,
                t.args,
                ast.TokenValue.RIGHT_BRACKETS,
            )
        if t.parent_rules:
            self.print(ast.TokenValue.LEFT_PARENTHESES)
            self.interleave(
                lambda: self.print(ast.TokenValue.COMMA, WHITESPACE),
                self.print,
                t.parent_rules,
            )
            self.print(ast.TokenValue.RIGHT_PARENTHESES)
        if t.for_host_name:
            self.print(
                WHITESPACE,
                ast.TokenValue.FOR,
                WHITESPACE,
                t.for_host_name,
            )
        self.print(
            ast.TokenValue.COLON,
            Indentation.IndentWithNewline,  # Schema Stmt indent
        )
        if t.doc:
            self.print(
                '"""',
                t.doc.replace('"', '\\"'),
                '"""',
                Indentation.Newline,
            )
        if t.checks:
            self.interleave(
                lambda: self.print(Indentation.Newline),
                self.print,
                t.checks,
            )
        self.write(NEWLINE)
        # Schema Stmt dedent
        self.print(Indentation.Dedent)

    def walk_Decorator(self, t: ast.Decorator):
        """ast.AST: Decorator

        Parameters
        ----------
        - name: Optional[Identifier]
        - args: Optional[CallExpr]
        """
        assert isinstance(t, ast.Decorator)
        self.print(
            ast.TokenValue.AT,
            t.name,
        )
        if t.args:
            self.print(
                ast.TokenValue.LEFT_PARENTHESES,
                t.args,
                ast.TokenValue.RIGHT_PARENTHESES,
            )
        self.writeln()

    def walk_Arguments(self, t: ast.Arguments):
        """ast.AST: Arguments

        Parameters
        ----------
        - args: List[Identifier] = []
        - type_annotation_list: List[str] = []
        - defaults: List[Expr] = []
        """
        assert isinstance(t, ast.Arguments)

        def write_argument(
            para: Tuple[ast.Identifier, Optional[str], Optional[ast.Expr]]
        ):
            arg, type_str, default = para
            self.print(
                arg,
                (": " + type_str) if type_str else "",
            )
            if default:
                self.print(WHITESPACE, ast.TokenValue.ASSIGN, WHITESPACE, default)

        self.interleave(
            lambda: self.write(COMMA_WHITESPACE),
            write_argument,
            zip(t.args, t.type_annotation_list, t.defaults),
        )

    def walk_SchemaAttr(self, t: ast.SchemaAttr):
        """ast.AST: SchemaAttr

        Parameters
        ----------
        - doc: str
        - name: str
        - type_str: str
        - is_optional: bool
        - value: Expr
        - decorators: List[Decorator]
        - op: Union[AugOp, Assign]
        """
        assert isinstance(t, ast.SchemaAttr)
        self.exprs(t.decorators)
        self.print(
            t.name,
            ast.TokenValue.QUESTION if t.is_optional else "",
        )
        self.print(ast.TokenValue.COLON, WHITESPACE, t.type_str)
        if t.op:
            self.print(
                WHITESPACE,
                ast.OPERATOR_VALUE_MAP[t.op],
                WHITESPACE,
                t.value,
            )
        self.write(NEWLINE)

    def walk_IfExpr(self, t: ast.IfExpr):
        """ast.AST: IfExpr

        Parameters
        ----------
        - cond: Expr
        - body: Expr
        - orelse: Expr
        """
        assert isinstance(t, ast.IfExpr)
        self.print(
            t.body,
            WHITESPACE,
            ast.TokenValue.IF,
            WHITESPACE,
            t.cond,
            WHITESPACE,
            ast.TokenValue.ELSE,
            WHITESPACE,
            t.orelse,
        )

    def walk_UnaryExpr(self, t: ast.UnaryExpr):
        """ast.AST: UnaryExpr(Expr)

        Parameters
        ----------
        - op: UnaryOp
        - operand: Expr
        """
        assert isinstance(t, ast.UnaryExpr)
        self.print(
            ast.OPERATOR_VALUE_MAP[t.op],
            WHITESPACE if t.op == ast.UnaryOp.Not else "",
            t.operand,
        )

    def walk_BinaryExpr(self, t: ast.BinaryExpr):
        """ast.AST: BinaryExpr

        Parameters
        ----------
        - left: Expr
        - right: Expr
        - op: BinaryOperator
        """
        assert isinstance(t, ast.BinaryExpr) and t.left and t.right and t.op
        self.print(
            t.left,
            WHITESPACE,
            ast.OPERATOR_VALUE_MAP[t.op],
            WHITESPACE,
            t.right,
        )

    def walk_SelectorExpr(self, t: ast.SelectorExpr):
        """ast.AST: SelectorExpr

        Parameters
        ----------
        - value: Expr
        - attr: Identifier
        - ctx: ExprContext
        - has_question: bool
        """
        assert isinstance(t, ast.SelectorExpr)
        self.print(
            t.value,
            ast.TokenValue.QUESTION if t.has_question else "",
            ast.TokenValue.DOT,
            t.attr,
        )

    def walk_CallExpr(self, t: ast.CallExpr):
        """ast.AST: CallExpr

        Parameters
        ----------
        - func: Expr
        - args: List[Expr]
        - keywords: List[Keyword]
        """
        assert isinstance(t, ast.CallExpr)
        if t.func:
            self.print(
                t.func,
                ast.TokenValue.LEFT_PARENTHESES,
            )
        self.write_args_and_kwargs(t.args, t.keywords)
        if t.func:
            self.print(
                ast.TokenValue.RIGHT_PARENTHESES,
            )

    def walk_ParenExpr(self, t: ast.ParenExpr):
        """ast.AST: ParenExpr

        Parameters
        ----------
        - expr: Expr
        """
        assert isinstance(t, ast.ParenExpr)
        self.print(
            ast.TokenValue.LEFT_PARENTHESES,
            t.expr,
            ast.TokenValue.RIGHT_PARENTHESES,
        )

    def walk_ListExpr(self, t: ast.ListExpr):
        """ast.AST: ListExpr

        Parameters
        ----------
        - elts: List[Expr]
        """
        assert isinstance(t, ast.ListExpr)
        in_one_line = len(set(map(lambda e: e.line, t.elts))) <= 1
        if len(t.elts or []) > 0:
            if isinstance(t.elts[0], ast.ListIfItemExpr):
                in_one_line = False
        self.write(ast.TokenValue.LEFT_BRACKETS)
        if t.elts:
            self.print(Indentation.IndentWithNewline if not in_one_line else "")
            splits = COMMA_WHITESPACE if in_one_line else Indentation.Newline
            self.interleave(
                lambda: self.print(splits),
                self.expr,
                t.elts,
            )
            self.print(Indentation.DedentWithNewline if not in_one_line else "")
        self.write(ast.TokenValue.RIGHT_BRACKETS)

    def walk_StarredExpr(self, t: ast.StarredExpr):
        assert isinstance(t, ast.StarredExpr) and t.value
        self.print(
            ast.TokenValue.MULTIPLY,
            t.value,
        )

    def walk_ListComp(self, t: ast.ListComp):
        """ast.AST: ListComp

        Parameters
        ----------
        - elt: Expr
        - generators: List[CompClause]
            - targets: List[Expr]
            - iter: Expr
            - ifs: List[Expr]
        """
        assert isinstance(t, ast.ListComp)
        self.write(ast.TokenValue.LEFT_BRACKETS)
        self.expr(t.elt)
        for gen in t.generators:
            self.expr(gen)
        self.write(ast.TokenValue.RIGHT_BRACKETS)

    def walk_DictComp(self, t: ast.DictComp):
        """ast.AST: DictComp

        Parameters
        ----------
        - key: Expr
        - value: Expr
        - generators: List[CompClause]
        """
        assert isinstance(t, ast.DictComp)
        self.write(ast.TokenValue.LEFT_BRACE)
        self.print(
            t.key,
            ast.TokenValue.COLON,
            WHITESPACE,
            t.value,
        )
        for gen in t.generators:
            self.expr(gen)
        self.write(ast.TokenValue.RIGHT_BRACE)

    def walk_CompClause(self, t: ast.CompClause):
        """ast.AST: CompClause

        Parameters
        ----------
        - targets: List[Expr]
        - iter: Expr
        - ifs: List[Expr]
        """
        assert isinstance(t, ast.CompClause)
        self.write_with_spaces(ast.TokenValue.FOR)
        self.interleave(lambda: self.write(COMMA_WHITESPACE), self.expr, t.targets)
        self.write_with_spaces(ast.TokenValue.IN)
        self.expr(t.iter)
        for if_clause in t.ifs:
            self.write_with_spaces(ast.TokenValue.IF)
            self.expr(if_clause)

    def walk_QuantExpr(self, t: ast.QuantExpr):
        """ast.AST: QuantExpr

        Parameters
        ----------
        - target: Optional[Expr] = None
        - variables: List[Identifier] = []
        - op: Optional[int] = None
        - test: Optional[Expr] = None
        - if_cond: Optional[Expr] = None
        - ctx: ExprContext = ctx
        """
        in_one_line = t.test.line == t.line
        self.print(
            QUANT_OP_TOKEN_VAL_MAPPING[t.op],
            WHITESPACE,
        )
        self.interleave(lambda: self.write(COMMA_WHITESPACE), self.expr, t.variables)
        self.write_with_spaces(ast.TokenValue.IN)
        self.expr(t.target)
        self.write(WHITESPACE)
        self.write(ast.TokenValue.LEFT_BRACE)
        self.print(Indentation.IndentWithNewline if not in_one_line else "")
        self.expr(t.test)
        if t.if_cond:
            self.print(WHITESPACE, ast.TokenValue.IF, WHITESPACE, t.if_cond)
        self.print(Indentation.DedentWithNewline if not in_one_line else "")
        self.write(ast.TokenValue.RIGHT_BRACE)

    def walk_Subscript(self, t: ast.Subscript):
        """ast.AST: Subscript

        Parameters
        ----------
        - value: Optional[Expr] = None
        - index: Optional[Expr] = None
        - lower: Optional[Expr] = None
        - upper: Optional[Expr] = None
        - step: Optional[Expr] = None
        - ctx: ExprContext = ExprContext.LOAD
        - has_question: bool = False
        """
        assert isinstance(t, ast.Subscript)
        self.print(
            t.value,
            ast.TokenValue.QUESTION if t.has_question else "",
            ast.TokenValue.LEFT_BRACKETS,
        )
        if t.index:
            self.expr(t.index)
        else:
            self.print(
                t.lower,
                ast.TokenValue.COLON,
                t.upper,
                ast.TokenValue.COLON,
                t.step,
            )
        self.print(ast.TokenValue.RIGHT_BRACKETS)

    def walk_SchemaExpr(self, t: ast.SchemaExpr):
        """ast.AST: SchemaExpr

        Parameters
        ----------
        - name: Identifier
        - config: ConfigExpr
        - args: List[Expr] = []
        - kwargs: List[Keyword] = []
        """
        assert isinstance(t, ast.SchemaExpr)
        self.print(t.name)

        if t.args or t.kwargs:
            self.write(ast.TokenValue.LEFT_PARENTHESES)
            self.write_args_and_kwargs(t.args, t.kwargs)
            self.write(ast.TokenValue.RIGHT_PARENTHESES)

        self.print(WHITESPACE, t.config)

    def walk_ConfigExpr(self, t: ast.ConfigExpr):
        """ast.AST: ConfigExpr

        Parameters
        ----------
        - items: List[ConfigEntry] = []
            - key: Expr = key
            - value: Expr = value
            - operation: int = ConfigEntryOperation.UNION
            - insert_index: Union[int, str] = -1
        """

        def write_config_key(key: ast.AST) -> int:
            """Write config key and return need right brace"""
            if isinstance(key, ast.Identifier):
                self.write_ast_comments(key)
                names = key.names
                # Judge contains string identifier, e.g., "x-y-z"
                need_right_brace = not all(
                    [bool(re.match(IDENTIFIER_REGEX, n)) for n in names]
                )
                if need_right_brace:
                    # a: {b: {c op value}}
                    self.print(
                        ": {".join(
                            ['"{}"'.format(n.replace('"', '\\"')) for n in names]
                        )
                    )
                    return len(names) - 1
                else:
                    # a.b.c op value
                    self.expr(key)
                    return 0
            else:
                self.expr(key)
                return 0

        def write_item(item: ast.ConfigEntry):
            if item.key is None:
                # for dictionary unpacking operator in dicts {**{'y': 2}}
                # see PEP 448 for details
                if not isinstance(item.value, ast.ConfigIfEntryExpr):
                    self.print(ast.TokenValue.DOUBLE_STAR)
                self.print(item.value)
            else:
                tok = ast.TokenValue.COLON
                if item.operation == ast.ConfigEntryOperation.INSERT:
                    tok = ast.TokenValue.COMP_PLUS
                elif item.operation == ast.ConfigEntryOperation.OVERRIDE:
                    tok = ast.TokenValue.ASSIGN
                print_right_brace_count = write_config_key(item.key)
                if item.insert_index is not None and item.insert_index >= 0:
                    self.print(
                        ast.TokenValue.LEFT_BRACKETS,
                        item.insert_index,
                        ast.TokenValue.RIGHT_BRACKETS,
                    )
                if tok != ast.TokenValue.COLON:
                    self.print(WHITESPACE)
                self.print(
                    tok,
                    WHITESPACE,
                    item.value,
                )
                self.print(ast.TokenValue.RIGHT_BRACE * (print_right_brace_count or 0))

        in_one_line = len(set(map(lambda e: e.line, t.items)).union([t.line])) == 1
        self.write(ast.TokenValue.LEFT_BRACE)
        if t.items:
            self.print(Indentation.IndentWithNewline if not in_one_line else "")
            self.interleave(
                lambda: self.print(COMMA_WHITESPACE) if in_one_line else self.writeln(),
                write_item,
                t.items,
            )
            self.print(Indentation.DedentWithNewline if not in_one_line else "")
        self.write(ast.TokenValue.RIGHT_BRACE)

    def walk_CheckExpr(self, t: ast.CheckExpr):
        """ast.AST: CheckExpr

        Parameters
        ----------
        - test: Expr
        - if_cond: Expr
        - msg: Expr
        """
        assert isinstance(t, ast.CheckExpr) and t.test
        self.expr(t.test)
        if t.if_cond:
            self.print(WHITESPACE, ast.TokenValue.IF, WHITESPACE, t.if_cond)
        if t.msg:
            self.print(
                COMMA_WHITESPACE,
                t.msg,
            )

    def walk_LambdaExpr(self, t: ast.LambdaExpr):
        """ast.AST: LambdaExpr

        Parameters
        ----------
        - args: Optional[Arguments]
        - return_type_str: Optional[str]
        - return_type_node: Optional[Type]
        - body: List[Stmt]
        """
        self.print(ast.TokenValue.LAMBDA)
        if t.args:
            self.print(
                WHITESPACE,
                t.args,
            )
        if t.return_type_str:
            self.print(
                WHITESPACE,
                ast.TokenValue.RIGHT_ARROW,
                WHITESPACE,
                t.return_type_str,
            )
        self.print(
            WHITESPACE,
            ast.TokenValue.LEFT_BRACE,
            NEWLINE,
            Indentation.Indent,
        )
        self.stmts(t.body)
        self.print(
            Indentation.Dedent,
            NEWLINE,
            ast.TokenValue.RIGHT_BRACE,
        )

    def walk_Compare(self, t: ast.Compare):
        """ast.AST: Compare

        Parameters
        ----------
        - left: Optional[Expr] = None
        - ops: List[CmpOp] = []
        - comparators: List[Expr] = []
        """
        assert isinstance(t, ast.Compare)
        self.expr(t.left)
        for op, expr in zip(t.ops, t.comparators):
            self.print(
                WHITESPACE,
                ast.OPERATOR_VALUE_MAP[op],
                WHITESPACE,
                expr,
            )

    def walk_Identifier(self, t: ast.Identifier):
        """ast.AST: Identifier

        Parameters
        ----------
        - names: List[str]
        """
        assert isinstance(t, ast.Identifier) and t.ctx
        # Convert pkgpath qualified name to a normal identifier
        if t.names[0].startswith("@"):
            pkgpath = t.names[0][1:]
            t.names[0] = self.import_spec.get(pkgpath) or pkgpath
        self.write(t.get_name())

    def walk_NumberLit(self, t: ast.AST):
        """ast.AST: NumberLit

        Parameters
        ----------
        - value: int
        """
        assert isinstance(t, ast.NumberLit)
        self.write(str(t.value))

    def walk_StringLit(self, t: ast.StringLit):
        """ast.AST: StringLit

        Parameters
        ----------
        - value: str
        - raw_value: str
        - is_long_string = False
        """
        assert isinstance(t, ast.StringLit)
        self.write(
            t.raw_value
            or (
                '"""{}"""'.format(t.value.replace('"', '\\"'))
                if t.is_long_string
                else '"{}"'.format(t.value.replace('"', '\\"'))
            )
        )

    def walk_NameConstantLit(self, t: ast.NameConstantLit):
        """ast.AST: NameConstantLit

        Parameters
        ----------
        - value
        """
        assert isinstance(t, ast.NameConstantLit)
        # None, Undefined, True, False
        self.write(str(t.value))

    def walk_JoinedString(self, t: ast.JoinedString):
        """ast.AST: JoinedString

        Parameters
        ----------
        - values: List[Union[Expr, StringLit]]

        TOS
        ---
        - format_spec
        - formatted expr list

        Operand
        -------
        Formatted expr list count
        """
        assert isinstance(t, ast.JoinedString)
        assert t.values
        quote_str = '"""' if t.is_long_string else '"'
        self.print(quote_str)
        for value in t.values:
            if isinstance(value, ast.FormattedValue):
                self.print(
                    "$",
                    ast.TokenValue.LEFT_BRACE,
                    value.value,
                )
                if value.format_spec:
                    self.print(
                        ast.TokenValue.COLON,
                        WHITESPACE,
                        value.format_spec,
                    )
                self.print(
                    ast.TokenValue.RIGHT_BRACE,
                )
            elif isinstance(value, ast.StringLit):
                self.write(
                    value.raw_value or "{}".format(value.value.replace('"', '\\"'))
                )
            elif isinstance(value, ast.Expr):
                self.expr(value)
            else:
                raise Exception("Invalid AST JoinedString children")
        self.print(quote_str)

    def walk_TypeAliasStmt(self, t: ast.TypeAliasStmt):
        """ast.AST: TypeAliasStmt

        Parameters
        ----------
        - type_name: Identifier
        - type_value: Type
        """
        self.print(
            ast.TokenValue.TYPE,
            WHITESPACE,
            t.type_name,
            WHITESPACE,
            ast.TokenValue.ASSIGN,
            WHITESPACE,
            t.type_value.plain_type_str,
            NEWLINE,
        )

    def walk_UnificationStmt(self, t: ast.UnificationStmt):
        """ast.AST: UnificationStmt

        Parameters
        ----------
        - target: Identifier
        - value: SchemaExpr
        """
        self.print(t.target, ast.TokenValue.COLON, WHITESPACE, t.value, NEWLINE)

    def walk_AssignStmt(self, t: ast.AssignStmt):
        """ast.AST: AssignStmt

        Parameters
        ----------
        - targets: List[Identifier]
        - value: Expr
        """
        assert isinstance(t, ast.AssignStmt) and t.targets and t.value
        for i, target in enumerate(t.targets):
            self.print(target)
            if i == 0 and t.type_annotation:
                self.print(ast.TokenValue.COLON, WHITESPACE, t.type_annotation)
            self.print(WHITESPACE, ast.TokenValue.ASSIGN, WHITESPACE)
        self.print(t.value, NEWLINE)
        if isinstance(t.value, ast.SchemaExpr):
            self.print(NEWLINE)

    def walk_AugAssignStmt(self, t: ast.AugAssignStmt):
        """ast.AST: AugAssignStmt

        Parameters
        ----------
        - target: Identifier
        - value: Expr
        - op: AugOp
        """
        assert isinstance(t, ast.AugAssignStmt) and t.target and t.value and t.op
        self.print(
            t.target,
            WHITESPACE,
            ast.OPERATOR_VALUE_MAP[t.op],
            WHITESPACE,
            t.value,
            NEWLINE,
        )

    def write_args_and_kwargs(self, args: List[ast.Expr], keywords: List[ast.Keyword]):
        def print_arg_assign_value(keyword: ast.Keyword):
            self.print(
                keyword.arg,
                ast.TokenValue.ASSIGN,
                keyword.value,
            )

        self.interleave(lambda: self.write(COMMA_WHITESPACE), self.expr, args)
        if args and keywords:
            self.print(COMMA_WHITESPACE)
        self.interleave(
            lambda: self.write(COMMA_WHITESPACE), print_arg_assign_value, keywords
        )

    def walk_ListIfItemExpr(self, t: ast.ListIfItemExpr):
        assert isinstance(t, ast.ListIfItemExpr)
        self.print(
            ast.TokenValue.IF,
            WHITESPACE,
            t.if_cond,
            ast.TokenValue.COLON,
            Indentation.IndentWithNewline,
        )
        self.interleave(lambda: self.print(NEWLINE), self.print, t.exprs)
        self.print(Indentation.DedentWithNewline)
        if t.orelse:
            if isinstance(t.orelse, ast.ListIfItemExpr):
                self.print("el")
                self.expr(t.orelse)
            elif isinstance(t.orelse, ast.ListExpr):
                self.print(
                    ast.TokenValue.ELSE,
                    ast.TokenValue.COLON,
                    Indentation.IndentWithNewline,
                )
                self.interleave(lambda: self.print(NEWLINE), self.print, t.orelse.elts)
                self.print(Indentation.Dedent)

    def walk_ConfigIfEntryExpr(self, t: ast.ConfigIfEntryExpr):
        assert isinstance(t, ast.ConfigIfEntryExpr)
        self.print(
            ast.TokenValue.IF,
            WHITESPACE,
            t.if_cond,
            ast.TokenValue.COLON,
            Indentation.IndentWithNewline,
        )

        def write_item(item):
            key, value, operation = item
            if key is None:
                # for dictionary unpacking operator in dicts {**{'y': 2}}
                # see PEP 448 for details
                if not isinstance(value, ast.ConfigIfEntryExpr):
                    self.print(ast.TokenValue.DOUBLE_STAR)
                self.print(value)
            else:
                tok = ast.TokenValue.COLON
                if operation == ast.ConfigEntryOperation.INSERT:
                    tok = ast.TokenValue.COMP_PLUS
                elif operation == ast.ConfigEntryOperation.OVERRIDE:
                    tok = ast.TokenValue.ASSIGN
                self.print(key)
                if tok != ast.TokenValue.COLON:
                    self.print(WHITESPACE)
                self.print(
                    tok,
                    WHITESPACE,
                    value,
                )

        self.interleave(
            lambda: self.writeln(),
            write_item,
            zip(t.keys, t.values, t.operations),
        )
        self.print(Indentation.DedentWithNewline)
        if t.orelse:
            if isinstance(t.orelse, ast.ConfigIfEntryExpr):
                self.print("el")
                self.expr(t.orelse)
            elif isinstance(t.orelse, ast.ConfigExpr):
                self.print(
                    ast.TokenValue.ELSE,
                    ast.TokenValue.COLON,
                    Indentation.IndentWithNewline,
                )
                self.interleave(
                    lambda: self.print(Indentation.Newline),
                    write_item,
                    zip(t.orelse.keys, t.orelse.values, t.orelse.operations),
                )
                self.print(Indentation.Dedent)


def PrintAST(
    node: ast.AST,
    out: Union[io.TextIOBase, io.StringIO] = sys.stdout,
    config: Config = Config(),
):
    """Print a KCL AST Module to `out` io with `config`"""
    Printer(config, out).print_ast(node)



# Copyright 2021 The KCL Authors. All rights reserved.

from dataclasses import dataclass, field
from typing import List, Optional
from enum import IntEnum


class Opcode(IntEnum):
    """Opcode class is used in KCL virtual machine.

    Virtual machine operator code can be divided into two categories
    - Stack opcode
    - Semantics opcode
    """

    INVALID = -1  # Invalid opcode

    POP_TOP = 1  # Stack pops 1 operand from the top of the stack
    ROT_TWO = 2  # Stack rotates 2 operands
    ROT_THREE = 3  # Stack rotates 3 operands
    DUP_TOP = 4  # Copy the reference of the top element of the Stack to a copy and put it on the top of the stack
    DUP_TOP_TWO = 5  # Copy the references of the first two elements on the top of the Stack to one copy and put it on the top of the stack.
    COPY_TOP = 6  # Copy the value of the top element of the Stack to a copy and put it on the top of the stack
    NOP = 9  # No operation

    UNARY_POSITIVE = 10  # +a
    UNARY_NEGATIVE = 11  # -a
    UNARY_NOT = 12  # not a
    UNARY_INVERT = 13  # ~a

    MEMBER_SHIP_AS = 16  # a as b

    BINARY_POWER = 20  # a ** b
    BINARY_MULTIPLY = 21  # a * b
    BINARY_MODULO = 22  # a % b
    BINARY_ADD = 23  # a + b
    BINARY_SUBTRACT = 24  # a - b
    BINARY_SUBSCR = 25  # b[a]
    BINARY_FLOOR_DIVIDE = 26  # a // b
    BINARY_TRUE_DIVIDE = 27  # a / b
    BINARY_LSHIFT = 28  # `a << b`
    BINARY_RSHIFT = 29  # `a >> b`
    BINARY_AND = 30  # `a & b`
    BINARY_XOR = 31  # `a ^ b`
    BINARY_OR = 32  # `a | b`
    BINARY_LOGIC_AND = 33  # `a and b`
    BINARY_LOGIC_OR = 34  # `a or b`

    INPLACE_FLOOR_DIVIDE = 40  # a //= b
    INPLACE_TRUE_DIVIDE = 41  # a /= b
    INPLACE_ADD = 42  # a += b
    INPLACE_SUBTRACT = 43  # `a -= b`
    INPLACE_MULTIPLY = 44  # `a *= b`
    INPLACE_MODULO = 45  # a %= b
    INPLACE_POWER = 46  # a **= b
    INPLACE_LSHIFT = 47  # a <<= b
    INPLACE_RSHIFT = 48  # a >>= b
    INPLACE_AND = 49  # a &= b
    INPLACE_XOR = 50  # a ^= b
    INPLACE_OR = 51  # a |= b

    COMPARE_EQUAL_TO = 60  # a == b
    COMPARE_NOT_EQUAL_TO = 61  # a != b
    COMPARE_LESS_THAN = 62  # a < b
    COMPARE_LESS_THAN_OR_EQUAL_TO = 63  # a <= b
    COMPARE_GREATER_THAN = 64  # a > b
    COMPARE_GREATER_THAN_OR_EQUAL_TO = 65  # a >= b
    COMPARE_IS = 66  # a is b
    COMPARE_IS_NOT = 67  # a is not b
    COMPARE_IN = 68  # a in b
    COMPARE_NOT_IN = 69  # a not in b

    STORE_MAP = 70  # Put a dict entry into the dict object
    STORE_SUBSCR = 71  # Put a subscript of collection
    DELETE_SUBSCR = 72  # Delete a subscript of collection
    BUILD_SCHEMA_CONFIG = 73  # Generate an empty schema config object
    STORE_SCHEMA_CONFIG = 74  # Put a schema config entry into the schema config object

    PRINT_EXPR = 80  # Print expression to stdout
    EMIT_EXPR = 81  # Emit a schema expression to the output

    SCHEMA_NOP = 90  # Expressions in the schema interval operation
    RETURN_VALUE = 91  # Return value in the schema
    RETURN_LAST_VALUE = 92  # Return the last value in the lambda expression

    HAVE_ARGUMENT = 99  # Opcodes from here have an argument

    STORE_NAME = 100  # Index in name list
    UNPACK_SEQUENCE = 101  # Number of sequence items
    GET_ITER = 102  # Get a element from a iterator
    FOR_ITER = 103  # Get a iterator from str/list/dict/schema
    STORE_ATTR = 105  # Index in name list
    DELETE_ATTR = 106  # Delete a attribute
    STORE_GLOBAL = 107  # Store a global variable
    DELETE_GLOBAL = 108  # Delete a global variable
    LOAD_GLOBAL = 109  # Index in name list
    LOAD_CONST = 110  # Index in const list
    LOAD_NAME = 111  # Index in name list
    LOAD_LOCAL = 112  # Local variable number
    STORE_LOCAL = 113  # Local variable number
    DELETE_LOCAL = 114  # Local variable number
    LOAD_FREE = 115  # Load from closure cell
    STORE_FREE = 116  # Store into cell
    DELETE_FREE = 117  # Delete closure cell
    BUILD_TUPLE = 118  # Number of tuple items
    BUILD_LIST = 119  # Number of list items
    BUILD_SET = 120  # Number of set items
    BUILD_MAP = 121  # Always zero for now
    LOAD_ATTR = 123  # Index in name list
    LOAD_BUILT_IN = 124  # Index in built-in list
    IMPORT_NAME = 126  # Index in name list
    COMPARE_OP = 127  # Comparison operator

    JUMP_FORWARD = 130  # Number of bytes to skip
    JUMP_IF_FALSE_OR_POP = 131  # Target byte offset from beginning of code
    JUMP_IF_TRUE_OR_POP = 132  # Target byte offset from beginning of code
    JUMP_ABSOLUTE = 133  # Target byte offset from beginning of code
    POP_JUMP_IF_FALSE = 134  # Target byte offset from beginning of code
    POP_JUMP_IF_TRUE = 135  # Target byte offset from beginning of code

    CALL_FUNCTION = 140  # #args + (#kwargs<<8) CALL_FUNCTION_XXX opcodes defined below depend on this definition
    MAKE_FUNCTION = 141  # #defaults + #kwdefaults<<8 + #annotations<<16
    BUILD_SLICE = 142  # Number of items
    MAKE_CLOSURE = 143  # same as MAKE_FUNCTION
    LOAD_CLOSURE = 144  # Load free variable from closure
    RAISE_VARARGS = 145  # Number of raise arguments (1, 2 or 3)
    RAISE_CHECK = 146  # Expressions in the check block

    LIST_APPEND = 150  # Append a item into the list used in the comprehension
    SET_ADD = 151  # Append a item into the set used in the comprehension
    MAP_ADD = 152  # Append a item into the dict used in the comprehension
    DELETE_ITEM = 153  # Delete a item into the dict used in the filter expression

    MAKE_SCHEMA = 160  # Build schema construct function
    BUILD_SCHEMA = 161  # Build a schema instance
    LOAD_BUILD_SCHEMA = 162  # Load schema
    SCHEMA_ATTR = 163  # Declare a schema attribute
    SCHEMA_LOAD_ATTR = 164  # Load attribute in the schema
    SCHEMA_UPDATE_ATTR = 165  # Update attribute in the schema
    MAKE_DECORATOR = 166  # Build a decorator in the schema

    FORMAT_VALUES = 170  # Format value in the string interpolation

    PUSH_FRAME_LOCALS = 175
    POP_FRAME_LOCALS = 176

    DEBUG_STACK = 180  # Debug stack
    DEBUG_LOCALS = 181  # Debug VM locals
    DEBUG_GLOBALS = 182  # Debug VM globals
    DEBUG_NAMES = 183  # Debug VM names

    @staticmethod
    def has_arg(code: int) -> bool:
        return code > Opcode.HAVE_ARGUMENT


@dataclass
class Pos:
    number: int = 0
    pos: int = 0
    filename: str = None
    lineno: int = None
    colno: int = None

    def get_pos(self) -> int:
        return self.pos

    def get_number(self) -> int:
        return self.number

    def get_lineno(self) -> int:
        return self.lineno

    def set_lineno(self, lineno: int):
        self.lineno = lineno

    def set_pos(self, number: int, pos: int) -> bool:
        self.number = number
        old_pos = self.pos
        self.pos = pos
        return old_pos != pos


@dataclass
class Instruction(Pos):
    op: Opcode = None

    @staticmethod
    def size() -> int:
        return 1

    def output(self) -> List[int]:
        return [self.op]


@dataclass
class Label(Pos):
    @staticmethod
    def size() -> int:
        return 0

    def output(self) -> List[int]:
        return []

    def stack_effect(self):
        return 0


@dataclass
class InstructionWithArg(Instruction):
    arg: int = 0

    def byte(self, arg) -> int:
        return arg & 0xFF

    @staticmethod
    def size() -> int:
        return 4

    def output(self) -> List[int]:
        out = [
            self.byte(self.op),
            self.byte(self.arg),
            self.byte(self.arg >> 8),
            self.byte(self.arg >> 16),
        ]
        return out


@dataclass
class JumpRel(InstructionWithArg):
    dest: Label = field(default_factory=Label)

    def output(self) -> List[int]:
        self.arg = self.dest.get_pos()
        return super().output()


@dataclass
class JumpAbs(InstructionWithArg):
    dest: Label = field(default_factory=Label)

    def output(self) -> List[int]:
        self.arg = self.dest.get_pos()
        return super().output()


@dataclass
class EmittedInstruction:
    opcode: Opcode = Opcode.INVALID
    position: int = 0


@dataclass
class CompilationScope:
    instructions: list
    last_instruction: Optional[EmittedInstruction] = field(
        default_factory=EmittedInstruction
    )
    previous_instruction: Optional[EmittedInstruction] = field(
        default_factory=EmittedInstruction
    )



# Copyright 2021 The KCL Authors. All rights reserved.

from typing import List
from dataclasses import dataclass
from collections import defaultdict

import kclvm.kcl.error as kcl_error

from .vertex import Vertex


NAME_NONE_BUCKET_KEY = "$name_none"


@dataclass
class UnifierConfig:
    """The vertex unification config"""

    check_unique: bool = False
    override: bool = False


class Unifier:
    def __init__(self, config: UnifierConfig = UnifierConfig()):
        self.config: UnifierConfig = config

    def unify(self, vertex: Vertex) -> Vertex:
        """The vertex unification function"""
        if not vertex or not isinstance(vertex, Vertex):
            return vertex
        # Using bucket map to check unique/merge option and store values
        bucket = defaultdict(list)
        for v in vertex.adjs or []:
            self.append_vertex_into_bucket(bucket, v)
        # Using the following configuration meta (filename/line/column) with the same name
        # to override the previous configuration e.g., `stack config -> base config`
        for k, v_list in bucket.items():
            if v_list:
                for j in range(len(v_list)):
                    v_list[j].meta = v_list[-1].meta
                # Merge vertices in the vertex list the with the same name
                bucket[k] = self.merge_vertices(v_list)
        # Merge the vertex adjs
        vertex.adjs = sum(bucket.values(), [])
        return vertex

    def merge_vertices(self, vertices: List[Vertex]) -> List[Vertex]:
        """Merge a vertex list with same names"""
        if not vertices or not isinstance(vertices, list):
            return []
        vertex_list = []
        # Merge all adjs in vertex with the same name,
        # if the adjs is None, append it into the vertex list
        total_adjs = sum([v.adjs or [] for v in vertices], [])
        is_unified = False
        meta_names = []
        for v in vertices:
            # If there are vertices in the list without adjs, they may have the
            # conflict values and put them into the vertex list
            # and deal the value conflict in VM
            if v.adjs is None and v.node:
                vertex_list.append(v)
            elif not is_unified:
                v.adjs = total_adjs
                vertex_list.append(self.unify(v))
                is_unified = True
            if v.config_meta.name and v.config_meta.name not in meta_names:
                if v.option.is_override:
                    meta_names = [v.config_meta.name]
                else:
                    meta_names.append(v.config_meta.name)
                if len(meta_names) >= 2 and v.option.is_union:
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.CompileError_TYPE,
                        file_msgs=[
                            kcl_error.ErrFileMsg(
                                filename=v.meta.filename,
                                line_no=v.meta.line,
                                col_no=v.meta.column,
                            )
                        ],
                        arg_msg=f"conflict unification types between {meta_names[-1]} and {meta_names[0]}",
                    )
        # Return the merged the vertex list
        return vertex_list

    def append_vertex_into_bucket(self, bucket: dict, v: Vertex):
        """Append the vertex into the bucket map with unique and override option"""
        if not isinstance(bucket, dict) or not isinstance(v, Vertex):
            return
        # Check unique key error
        if self.config.check_unique and len(bucket[v.name]) > 1 and v.option.is_unique:
            kcl_error.report_exception(
                err_type=kcl_error.ErrType.CompileError_TYPE,
                file_msgs=[
                    kcl_error.ErrFileMsg(
                        filename=v.meta.filename,
                        line_no=v.node.line,
                        col_no=v.node.column,
                        end_col_no=v.node.end_column,
                    )
                ],
                arg_msg=f"config of '{v.name}' must be unique",
            )
        # Put the different value into the bucket with the same `name`.
        # Please note that then the vertex key is a runtime variable
        # such as string interpolation, the name is None e.g., `{"${name}": "Alice"}`
        # else the name is the key name such as `name` in `{"name": "Alice"}`
        if v.name is not None:
            # Override the value in the bucket
            if self.config.override:
                bucket[v.name] = [v]
            # Append the multiple values into the bucket and calculate these values in VM
            else:
                bucket[v.name].append(v)
        # Store the name none node including unpack expression and collection if expression
        else:
            bucket[NAME_NONE_BUCKET_KEY].append(v)



# Copyright 2021 The KCL Authors. All rights reserved.

import re
import typing
import copy
from collections import OrderedDict

import kclvm.kcl.error as kcl_error
import kclvm.kcl.info as kcl_info
import kclvm.kcl.ast as ast
import kclvm.compiler.parser.lark_parser as lark_parser
import kclvm.compiler.vfs as vfs

PKGPATH_IDENTIFIER_DOT_REGEX = r"[\d\w_]+\."
PKGPATH_DOT_REGEX = r"@[\d\w_\.]+\."


def _get_global_names(m: ast.Module) -> typing.List[str]:
    assert m
    assert isinstance(m, ast.Module)

    global_name_dict: typing.Dict[str, ast.AST] = OrderedDict()

    def walkFn_global(t: ast.AST) -> typing.Optional[typing.Callable]:
        nonlocal global_name_dict

        if isinstance(t, (ast.SchemaStmt, ast.RuleStmt)):
            node = t
            if kcl_info.isprivate_field(node.name) or node.name not in global_name_dict:
                global_name_dict[node.name] = node
            else:
                kcl_error.report_exception(
                    err_type=kcl_error.ErrType.UniqueKeyError_TYPE,
                    file_msgs=[
                        kcl_error.ErrFileMsg(
                            filename=lark_parser.filename,
                            line_no=node.line,
                            col_no=node.column,
                        )
                    ],
                    arg_msg=kcl_error.UNIQUE_KEY_MSG.format(node.name),
                )
            return None

        if isinstance(t, (ast.SchemaExpr, ast.LambdaExpr)):
            return None

        if isinstance(t, ast.ImportStmt):
            return None

        if isinstance(t, (ast.ListComp, ast.DictComp)):
            return None

        if isinstance(t, ast.AssignStmt):
            node = typing.cast(ast.AssignStmt, t)
            for expr in node.targets:
                if not isinstance(expr, ast.Identifier) or isinstance(
                    node.value, ast.LambdaExpr
                ):
                    continue

                ident = typing.cast(ast.Identifier, expr)
                is_config = isinstance(node.value, ast.SchemaExpr)
                if (
                    kcl_info.isprivate_field(ident.names[0])
                    or (ident.names[0] not in global_name_dict)
                    or is_config
                ):
                    global_name_dict[ident.names[0]] = node
                else:
                    kcl_error.report_exception(
                        err_type=kcl_error.ErrType.ImmutableCompileError_TYPE,
                        file_msgs=[
                            kcl_error.ErrFileMsg(
                                filename=lark_parser.filename,
                                line_no=ident.line,
                                col_no=ident.column,
                                end_col_no=ident.end_column,
                            )
                        ],
                    )

        # continue walk
        return walkFn_global

    # walk tree
    ast.WalkTree(m, walkFn_global)

    # dict to list
    return list(global_name_dict.keys())


def _get_schema_local_names(
    schema: typing.Union[ast.SchemaStmt, ast.RuleStmt]
) -> typing.List[str]:
    assert schema
    assert isinstance(schema, (ast.SchemaStmt, ast.RuleStmt))

    local_name_dict: typing.Dict[str, ast.AST] = OrderedDict()

    # walk args
    if schema.args:
        for x in schema.args.args:
            assert len(x.names) == 1, f"schema.args={schema.args}"
            local_name_dict[x.names[0]] = x

    def walkFn_schema_local(t: ast.AST) -> typing.Optional[typing.Callable]:
        if isinstance(t, ast.SchemaAttr):
            node = typing.cast(ast.SchemaAttr, t)
            local_name_dict[node.name] = node
            return None

        if isinstance(t, ast.AssignStmt):
            node = typing.cast(ast.AssignStmt, t)

            # a = b = c.d = value
            for expr in node.targets:
                if not isinstance(expr, ast.Identifier):
                    continue

                ident = typing.cast(ast.Identifier, expr)

                # skip: c.d = value
                if len(ident.names) == 1:
                    local_name_dict[ident.names[0]] = ident

        # continue walk
        return walkFn_schema_local

    # walk tree
    ast.WalkTree(schema, walkFn_schema_local)

    return list(local_name_dict.keys())


def fix_set_parent_info(m: ast.Module):
    """
    set parent info on ast
    :param m: target module ast
    """

    def _walk(t: ast.AST):
        def _set_parent(
            parent: ast.AST, inner: typing.Union[typing.List, typing.Dict, ast.AST]
        ):
            if isinstance(inner, list):
                [_set_parent(parent, item) for item in inner]
                return
            if isinstance(inner, dict):
                [_set_parent(parent, v) for _, v in inner]
                return
            if isinstance(inner, ast.AST):
                inner.parent = parent
                _walk(inner)

        for _, value in ast.iter_fields(t):
            _set_parent(t, value)

    assert m and isinstance(m, ast.Module)
    return _walk(m)


def fix_qualified_identifier(
    m: ast.Module, *, import_names: typing.Optional[typing.Dict[str, str]] = None
):
    """
    import path.to.pkg as pkgname

    x = pkgname.Name
    """
    # 0. init import names
    if import_names is None or not isinstance(import_names, dict):
        import_names = {}
    for import_spec in m.GetImportList():
        import_names[import_spec.name] = import_spec.path

    # 1. init global names
    _global_names = _get_global_names(m)
    for name in _global_names:
        if name not in m.global_names:
            m.global_names.append(name)

    # 2. init schema local name
    _schema_local_names: typing.Dict[str, typing.List[str]] = {}
    for schema in m.GetSchemaAndRuleList():
        _schema_local_names[schema.name] = _get_schema_local_names(schema)
        if schema.name not in m.local_names:
            m.local_names = _schema_local_names

    current_schema_name = ""
    generator_local_vars = []

    def walkFn_fix_global_ident(t: ast.AST) -> typing.Optional[typing.Callable]:
        if isinstance(t, (ast.DictComp, ast.ListComp)):
            for gen in t.generators or []:
                for ident in gen.targets:
                    generator_local_vars.append(ident.get_first_name())
                ast.WalkTree(gen, walkFn_fix_global_ident)
            if isinstance(t, ast.ListComp):
                ast.WalkTree(t.elt, walkFn_fix_global_ident)
            if isinstance(t, ast.DictComp):
                ast.WalkTree(t.key, walkFn_fix_global_ident)
                ast.WalkTree(t.value, walkFn_fix_global_ident)
            generator_local_vars.clear()
            return None
        elif isinstance(t, ast.QuantExpr):
            for ident in t.variables:
                generator_local_vars.append(ident.get_first_name())
            ast.WalkTree(t.target, walkFn_fix_global_ident)
            ast.WalkTree(t.test, walkFn_fix_global_ident)
            ast.WalkTree(t.if_cond, walkFn_fix_global_ident)
            generator_local_vars.clear()
        if not isinstance(t, ast.Identifier):
            return walkFn_fix_global_ident

        ident = typing.cast(ast.Identifier, t)
        if len(ident.names) < 2:
            return None

        # skip global name and generator local variables in list/dict comp and quant expression
        if ident.names[0] in _global_names or ident.names[0] in generator_local_vars:
            return None

        # fix qualified identifier
        if ident.names[0] in import_names:
            ident.pkgpath = import_names[ident.names[0]]

        return None

    def walkFn_fix_schema_ident(t: ast.AST) -> typing.Optional[typing.Callable]:
        nonlocal current_schema_name
        assert current_schema_name, f"current_schema_name={current_schema_name}"

        if not isinstance(t, ast.Identifier):
            return walkFn_fix_global_ident

        ident = typing.cast(ast.Identifier, t)
        if len(ident.names) < 2:
            return None

        # skip local name
        _local_names = _schema_local_names[current_schema_name]
        if ident.names[0] in _local_names:
            return None

        # skip global name
        if ident.names[0] in _global_names:
            return None

        # fix qualified identifier
        if ident.names[0] in import_names:
            ident.pkgpath = import_names[ident.names[0]]

        return None

    # -----------------------------------------------------

    # 3. fix all ident
    for stmt in m.body or []:
        if isinstance(stmt, (ast.SchemaStmt, ast.RuleStmt)):
            node = stmt
            current_schema_name = node.name
            ast.WalkTree(node, walkFn_fix_schema_ident)
            current_schema_name = ""
            continue

        ast.WalkTree(stmt, walkFn_fix_global_ident)

    # OK
    return


def fix_and_get_module_import_list(
    root: str, m: ast.Module, is_fix: bool = True, reversed: bool = False
) -> typing.List[ast.ImportStmt]:
    assert m
    assert isinstance(m, ast.Module)
    assert m.pkg

    import_spec_list: typing.List[ast.ImportStmt] = []
    pkgpath_table = {}

    for stmt in m.body or []:
        if not isinstance(stmt, ast.ImportStmt):
            continue

        if is_fix:
            assert stmt.path
            assert stmt.pkg_name

            stmt.rawpath = stmt.path

            stmt.path = vfs.FixImportPath(root, m.filename, stmt.path)
            stmt.name = stmt.pkg_name
        if reversed:
            pkgpath_table[stmt.path] = stmt.name
        else:
            pkgpath_table[stmt.name] = stmt.path

        import_spec = copy.deepcopy(stmt)
        import_spec_list.append(import_spec)

    if not is_fix:
        return import_spec_list

    # fix types name
    # asname.Name => @abs.pkg.Name
    # [asname.Name] => [@abs.pkg.Name]
    # {str:asname.Name} => {str:@abs.pkg.Name}
    # {str:[asname.Name]} => {str:[@abs.pkg.Name]}
    # asname1.Name1 | asname2.Name2 => @abs.pkg1.Name1 | @abs.pkg2.Name2
    for stmt in m.body or []:
        if isinstance(stmt, ast.AssignStmt):
            assign_stmt = typing.cast(ast.AssignStmt, stmt)
            if assign_stmt.type_annotation:
                if reversed:
                    assign_stmt.type_annotation = re.sub(
                        PKGPATH_DOT_REGEX,
                        lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                        if x.group()[1:-1] in pkgpath_table
                        else x.group(),
                        assign_stmt.type_annotation,
                    )
                else:
                    assign_stmt.type_annotation = re.sub(
                        PKGPATH_IDENTIFIER_DOT_REGEX,
                        lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                        if x.group()[:-1] in pkgpath_table
                        else x.group(),
                        assign_stmt.type_annotation,
                    )
        elif isinstance(stmt, ast.SchemaStmt):
            schema_stmt = typing.cast(ast.SchemaStmt, stmt)
            # Fix schema arguments type
            if schema_stmt.args and schema_stmt.args.type_annotation_list:
                for i, _type in enumerate(schema_stmt.args.type_annotation_list):
                    # if the `_type` is None, the schema argument has no any type annotation
                    if not _type:
                        continue
                    if reversed:
                        schema_stmt.args.type_annotation_list[i] = re.sub(
                            PKGPATH_DOT_REGEX,
                            lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                            if x.group()[1:-1] in pkgpath_table
                            else x.group(),
                            _type,
                        )
                    else:
                        schema_stmt.args.type_annotation_list[i] = re.sub(
                            PKGPATH_IDENTIFIER_DOT_REGEX,
                            lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                            if x.group()[:-1] in pkgpath_table
                            else x.group(),
                            _type,
                        )
            # Fix schame attr type
            for attr in schema_stmt.body or []:
                if not isinstance(attr, ast.SchemaAttr):
                    continue
                schema_attr = typing.cast(ast.SchemaAttr, attr)
                _type = schema_attr.type_str
                if reversed:
                    schema_attr.type_str = re.sub(
                        PKGPATH_DOT_REGEX,
                        lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                        if x.group()[1:-1] in pkgpath_table
                        else x.group(),
                        _type,
                    )
                else:
                    schema_attr.type_str = re.sub(
                        PKGPATH_IDENTIFIER_DOT_REGEX,
                        lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                        if x.group()[:-1] in pkgpath_table
                        else x.group(),
                        _type,
                    )
        elif isinstance(stmt, ast.RuleStmt):
            rule_stmt = typing.cast(ast.RuleStmt, stmt)
            if rule_stmt.args and rule_stmt.args.type_annotation_list:
                for i, _type in enumerate(rule_stmt.args.type_annotation_list):
                    # if the `_type` is None, the rule argument has no any type annotation
                    if not _type:
                        continue
                    if reversed:
                        rule_stmt.args.type_annotation_list[i] = re.sub(
                            PKGPATH_DOT_REGEX,
                            lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                            if x.group()[1:-1] in pkgpath_table
                            else x.group(),
                            _type,
                        )
                    else:
                        rule_stmt.args.type_annotation_list[i] = re.sub(
                            PKGPATH_IDENTIFIER_DOT_REGEX,
                            lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                            if x.group()[:-1] in pkgpath_table
                            else x.group(),
                            _type,
                        )
        elif isinstance(stmt, ast.TypeAliasStmt):
            # Fix rule arguments type
            type_alias_stmt = typing.cast(ast.TypeAliasStmt, stmt)
            if type_alias_stmt.type_value.plain_type_str:
                if reversed:
                    type_alias_stmt.type_value.plain_type_str = re.sub(
                        PKGPATH_DOT_REGEX,
                        lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                        if x.group()[1:-1] in pkgpath_table
                        else x.group(),
                        type_alias_stmt.type_value.plain_type_str,
                    )
                else:
                    type_alias_stmt.type_value.plain_type_str = re.sub(
                        PKGPATH_IDENTIFIER_DOT_REGEX,
                        lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                        if x.group()[:-1] in pkgpath_table
                        else x.group(),
                        type_alias_stmt.type_value.plain_type_str,
                    )

    class TypeNameTransformer(ast.TreeTransformer):
        def walk_LambdaExpr(self, node: ast.LambdaExpr):
            if node.args and node.args.type_annotation_list:
                for i, _type in enumerate(node.args.type_annotation_list):
                    # if the `_type` is None, the schema argument has no any type annotation
                    if not _type:
                        continue
                    if reversed:
                        node.args.type_annotation_list[i] = re.sub(
                            PKGPATH_DOT_REGEX,
                            lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                            if x.group()[1:-1] in pkgpath_table
                            else x.group(),
                            _type,
                        )
                    else:
                        node.args.type_annotation_list[i] = re.sub(
                            PKGPATH_IDENTIFIER_DOT_REGEX,
                            lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                            if x.group()[:-1] in pkgpath_table
                            else x.group(),
                            _type,
                        )
            if node.return_type_str:
                if reversed:
                    node.return_type_str = re.sub(
                        PKGPATH_DOT_REGEX,
                        lambda x: f"{pkgpath_table[x.group()[1:-1]]}."
                        if x.group()[1:-1] in pkgpath_table
                        else x.group(),
                        node.return_type_str,
                    )
                else:
                    node.return_type_str = re.sub(
                        PKGPATH_IDENTIFIER_DOT_REGEX,
                        lambda x: f"@{pkgpath_table[x.group()[:-1]]}."
                        if x.group()[:-1] in pkgpath_table
                        else x.group(),
                        node.return_type_str,
                    )
            return node

    TypeNameTransformer().walk(m)

    return import_spec_list


def fix_test_schema_auto_relaxed(m: ast.Module):
    if not m.filename.endswith("_test.k"):
        return

    for stmt in m.body or []:
        if not isinstance(stmt, ast.SchemaStmt):
            continue

        schema = typing.cast(ast.SchemaStmt, stmt)
        if schema.name.startswith("Test"):
            for x in schema.body or []:
                if not isinstance(x, ast.SchemaAttr):
                    continue
                attr = typing.cast(ast.SchemaAttr, x)
                attr.type_str = ""
                attr.is_optional = True

    return



# Copyright 2020 The KCL Authors. All rights reserved.

# Auto generated by {gen_lark_token.go & kcl.lark}; DO NOT EDIT!!!


class LarkToken:
    # kcl.lark rules and tokens (len=186)
    L_start = "start"  # start: (NEWLINE | statement)* ...
    L_statement = "statement"  # statement: simple_stmt | compound_stmt ...
    L_simple_stmt = "simple_stmt"  # simple_stmt: (assign_stmt | unification_stmt | expr_stmt | assert_stmt | import_stmt | type_alias_stmt) NEWLINE ...
    L_compound_stmt = (
        "compound_stmt"  # compound_stmt: if_stmt | schema_stmt | rule_stmt ...
    )
    L_import_stmt = "import_stmt"  # import_stmt: IMPORT dot_name (AS NAME)? ...
    L_dot_name = "dot_name"  # dot_name: (leading_dots identifier) | identifier ...
    L_leading_dots = "leading_dots"  # leading_dots: DOT+ ...
    L_assert_stmt = "assert_stmt"  # assert_stmt: ASSERT simple_expr (IF simple_expr)? (COMMA test)? ...
    L_if_stmt = "if_stmt"  # if_stmt: IF test COLON execution_block (ELIF test COLON execution_block)* (ELSE COLON execution_block)? ...
    L_execution_block = "execution_block"  # execution_block: if_simple_stmt | NEWLINE _INDENT schema_init_stmt+ _DEDENT ...
    L_if_simple_stmt = "if_simple_stmt"  # if_simple_stmt: (simple_assign_stmt | unification_stmt | expr_stmt | assert_stmt) NEWLINE ...
    L_assign_stmt = "assign_stmt"  # assign_stmt: identifier [COLON type] (ASSIGN identifier)* ASSIGN test ...
    L_simple_assign_stmt = (
        "simple_assign_stmt"  # simple_assign_stmt: identifier ASSIGN test ...
    )
    L_unification_stmt = (
        "unification_stmt"  # unification_stmt: identifier COLON schema_expr ...
    )
    L_schema_stmt = "schema_stmt"  # schema_stmt: [decorators] (SCHEMA|MIXIN|PROTOCOL) NAME [LEFT_BRACKETS [schema_arguments] RIGHT_BRACKETS] [LEFT_PARENTHESES identifier (COMMA identifier)* RIGHT_PARENTHESES] [for_host] COLON NEWLINE [schema_body] ...
    L_schema_arguments = "schema_arguments"  # schema_arguments: schema_argument (COMMA schema_argument)* ...
    L_schema_argument = (
        "schema_argument"  # schema_argument: NAME [COLON type] [ASSIGN test] ...
    )
    L_schema_body = "schema_body"  # schema_body: _INDENT (string NEWLINE)* [mixin_stmt] (schema_attribute_stmt|schema_init_stmt|schema_index_signature)* [check_block] _DEDENT ...
    L_schema_attribute_stmt = (
        "schema_attribute_stmt"  # schema_attribute_stmt: attribute_stmt NEWLINE ...
    )
    L_attribute_stmt = "attribute_stmt"  # attribute_stmt: [decorators] identifier [QUESTION] COLON type [(ASSIGN|COMP_OR) test] ...
    L_schema_init_stmt = (
        "schema_init_stmt"  # schema_init_stmt: if_simple_stmt | if_stmt ...
    )
    L_schema_index_signature = "schema_index_signature"  # schema_index_signature: LEFT_BRACKETS [NAME COLON] [ELLIPSIS] basic_type RIGHT_BRACKETS COLON type [ASSIGN test] NEWLINE ...
    L_rule_stmt = "rule_stmt"  # rule_stmt: [decorators] RULE NAME [LEFT_BRACKETS [schema_arguments] RIGHT_BRACKETS] [LEFT_PARENTHESES identifier (COMMA identifier)* RIGHT_PARENTHESES] [for_host] COLON NEWLINE [rule_body] ...
    L_rule_body = (
        "rule_body"  # rule_body: _INDENT (string NEWLINE)* check_expr+ _DEDENT ...
    )
    L_for_host = "for_host"  # for_host: FOR identifier ...
    L_decorators = "decorators"  # decorators: (AT decorator_expr NEWLINE)+ ...
    L_decorator_expr = "decorator_expr"  # decorator_expr: identifier [call_suffix] ...
    L_type = "type"  # @type: type_element (OR type_element)* ...
    L_type_element = "type_element"  # type_element: schema_type | basic_type | compound_type | literal_type ...
    L_schema_type = "schema_type"  # schema_type: identifier ...
    L_basic_type = "basic_type"  # basic_type: STRING_TYPE | INT_TYPE | FLOAT_TYPE | BOOL_TYPE | ANY_TYPE ...
    L_compound_type = "compound_type"  # compound_type: list_type | dict_type ...
    L_list_type = "list_type"  # list_type: LEFT_BRACKETS (type)? RIGHT_BRACKETS ...
    L_dict_type = (
        "dict_type"  # dict_type: LEFT_BRACE (type)? COLON (type)? RIGHT_BRACE ...
    )
    L_literal_type = (
        "literal_type"  # literal_type: string | number | TRUE | FALSE | NONE ...
    )
    L_type_alias_stmt = "type_alias_stmt"  # type_alias_stmt: TYPE NAME ASSIGN type ...
    L_check_block = "check_block"  # check_block: CHECK COLON NEWLINE _INDENT check_expr+ _DEDENT ...
    L_check_expr = "check_expr"  # check_expr: simple_expr [IF simple_expr] [COMMA primary_expr] NEWLINE ...
    L_mixin_stmt = "mixin_stmt"  # mixin_stmt: MIXIN LEFT_BRACKETS [mixins | multiline_mixins] RIGHT_BRACKETS NEWLINE ...
    L_multiline_mixins = "multiline_mixins"  # multiline_mixins: NEWLINE _INDENT mixins NEWLINE _DEDENT ...
    L_mixins = "mixins"  # mixins: identifier (COMMA (NEWLINE mixins | identifier))* ...
    L_expr_stmt = "expr_stmt"  # expr_stmt: testlist_expr ...
    L_testlist_expr = "testlist_expr"  # testlist_expr: test (COMMA test)* ...
    L_test = "test"  # test: if_expr | simple_expr ...
    L_if_expr = "if_expr"  # if_expr: simple_expr IF simple_expr ELSE test ...
    L_simple_expr = (
        "simple_expr"  # simple_expr: unary_expr | binary_expr | primary_expr ...
    )
    L_unary_expr = "unary_expr"  # unary_expr: un_op simple_expr ...
    L_binary_expr = "binary_expr"  # binary_expr: simple_expr bin_op simple_expr ...
    L_bin_op = "bin_op"  # bin_op: L_OR | L_AND ...
    L_un_op = "un_op"  # un_op: L_NOT | PLUS | MINUS | NOT ...
    L_primary_expr = "primary_expr"  # primary_expr: identifier call_suffix | operand | primary_expr select_suffix | primary_expr call_suffix | primary_expr slice_suffix ...
    L_operand = "operand"  # operand: identifier | number | string | constant | quant_expr | list_expr | list_comp | config_expr | dict_comp | schema_expr | lambda_expr | LEFT_PARENTHESES test RIGHT_PARENTHESES ...
    L_select_suffix = "select_suffix"  # select_suffix: [QUESTION] DOT NAME ...
    L_call_suffix = "call_suffix"  # call_suffix: LEFT_PARENTHESES [arguments [COMMA]] RIGHT_PARENTHESES ...
    L_slice_suffix = "slice_suffix"  # slice_suffix: [QUESTION] LEFT_BRACKETS (test | [test] COLON [test] [COLON [test]]) RIGHT_BRACKETS ...
    L_arguments = "arguments"  # arguments: argument (COMMA argument)* ...
    L_argument = "argument"  # argument: test | NAME ASSIGN test | MULTIPLY test | DOUBLE_STAR test ...
    L_identifier = "identifier"  # identifier: NAME (DOT NAME)* ...
    L_quant_expr = "quant_expr"  # quant_expr: quant_op [ identifier COMMA ] identifier IN quant_target LEFT_BRACE (simple_expr [IF simple_expr] | NEWLINE _INDENT simple_expr [IF simple_expr] NEWLINE _DEDENT)? RIGHT_BRACE ...
    L_quant_target = "quant_target"  # quant_target: string | identifier | list_expr | list_comp | config_expr | dict_comp ...
    L_quant_op = "quant_op"  # quant_op: ALL | ANY | FILTER | MAP ...
    L_list_expr = "list_expr"  # list_expr: LEFT_BRACKETS [list_items | NEWLINE [_INDENT list_items _DEDENT]] RIGHT_BRACKETS ...
    L_list_items = "list_items"  # list_items: list_item ((COMMA [NEWLINE] | [NEWLINE]) list_item)* [COMMA] [NEWLINE] ...
    L_list_item = "list_item"  # list_item: test | star_expr | if_item ...
    L_list_comp = "list_comp"  # list_comp: LEFT_BRACKETS (list_item comp_clause+ | NEWLINE _INDENT list_item comp_clause+ _DEDENT) RIGHT_BRACKETS ...
    L_dict_comp = "dict_comp"  # dict_comp: LEFT_BRACE (entry comp_clause+ | NEWLINE _INDENT entry comp_clause+ _DEDENT) RIGHT_BRACE ...
    L_entry = "entry"  # entry: test (COLON | ASSIGN | COMP_PLUS) test ...
    L_comp_clause = "comp_clause"  # comp_clause: FOR loop_variables [COMMA] IN simple_expr [NEWLINE] [IF test [NEWLINE]] ...
    L_if_entry = "if_entry"  # if_entry: IF test COLON if_entry_exec_block (ELIF test COLON if_entry_exec_block)* (ELSE COLON if_entry_exec_block)? ...
    L_if_entry_exec_block = "if_entry_exec_block"  # if_entry_exec_block: (test (COLON | ASSIGN | COMP_PLUS) test | double_star_expr | if_entry) [NEWLINE] | NEWLINE _INDENT (test (COLON | ASSIGN | COMP_PLUS) test | double_star_expr | if_entry) ((COMMA [NEWLINE] | [NEWLINE]) (test (COLON | ASSIGN | COMP_PLUS) test | double_star_expr | if_entry))* [COMMA] [NEWLINE] _DEDENT ...
    L_if_item = "if_item"  # if_item: IF test COLON if_item_exec_block (ELIF test COLON if_item_exec_block)* (ELSE COLON if_item_exec_block)? ...
    L_if_item_exec_block = "if_item_exec_block"  # if_item_exec_block: list_item [NEWLINE] | NEWLINE _INDENT list_item ((COMMA [NEWLINE] | NEWLINE) list_item)* [COMMA] [NEWLINE] _DEDENT ...
    L_star_expr = "star_expr"  # star_expr: MULTIPLY primary_expr ...
    L_double_star_expr = (
        "double_star_expr"  # double_star_expr: DOUBLE_STAR primary_expr ...
    )
    L_loop_variables = (
        "loop_variables"  # loop_variables: primary_expr (COMMA primary_expr)* ...
    )
    L_schema_expr = "schema_expr"  # schema_expr: identifier (LEFT_PARENTHESES [arguments] RIGHT_PARENTHESES)? config_expr ...
    L_config_expr = "config_expr"  # config_expr: LEFT_BRACE [config_entries | NEWLINE [_INDENT config_entries _DEDENT]] RIGHT_BRACE ...
    L_config_entries = "config_entries"  # config_entries: config_entry ((COMMA [NEWLINE] | [NEWLINE]) config_entry)* [COMMA] [NEWLINE] ...
    L_config_entry = "config_entry"  # config_entry: test (COLON | ASSIGN | COMP_PLUS) test | double_star_expr | if_entry ...
    L_lambda_expr = "lambda_expr"  # lambda_expr: LAMBDA [schema_arguments] [RIGHT_ARROW type] LEFT_BRACE [expr_stmt | NEWLINE _INDENT schema_init_stmt+ _DEDENT] RIGHT_BRACE ...
    L_number = "number"  # number: DEC_NUMBER [multiplier] | HEX_NUMBER | BIN_NUMBER | OCT_NUMBER | FLOAT_NUMBER ...
    L_multiplier = "multiplier"  # multiplier: SI_N_L | SI_U_L | SI_M_L | SI_K_L | SI_K | SI_M | SI_G | SI_T | SI_P  ...
    L_string = "string"  # string: STRING | LONG_STRING ...
    L_ASSIGN = "ASSIGN"  # ASSIGN: "=" ...
    L_COLON = "COLON"  # COLON: ":" ...
    L_SEMI_COLON = "SEMI_COLON"  # SEMI_COLON: ";" ...
    L_COMMA = "COMMA"  # COMMA: "," ...
    L_QUESTION = "QUESTION"  # QUESTION: "?" ...
    L_ELLIPSIS = "ELLIPSIS"  # ELLIPSIS: "..." ...
    L_RIGHT_ARROW = "RIGHT_ARROW"  # RIGHT_ARROW: "->" ...
    L_LEFT_PARENTHESES = "LEFT_PARENTHESES"  # LEFT_PARENTHESES: "(" ...
    L_RIGHT_PARENTHESES = "RIGHT_PARENTHESES"  # RIGHT_PARENTHESES: ")" ...
    L_LEFT_BRACKETS = "LEFT_BRACKETS"  # LEFT_BRACKETS: "[" ...
    L_RIGHT_BRACKETS = "RIGHT_BRACKETS"  # RIGHT_BRACKETS: "]" ...
    L_LEFT_BRACE = "LEFT_BRACE"  # LEFT_BRACE: "{" ...
    L_RIGHT_BRACE = "RIGHT_BRACE"  # RIGHT_BRACE: "}" ...
    L_PLUS = "PLUS"  # PLUS: "+" ...
    L_MINUS = "MINUS"  # MINUS: "-" ...
    L_MULTIPLY = "MULTIPLY"  # MULTIPLY: "*" ...
    L_DIVIDE = "DIVIDE"  # DIVIDE: "/" ...
    L_MOD = "MOD"  # MOD: "%" ...
    L_DOT = "DOT"  # DOT: "." ...
    L_AND = "AND"  # AND: "&" ...
    L_OR = "OR"  # OR: "|" ...
    L_XOR = "XOR"  # XOR: "^" ...
    L_NOT = "NOT"  # NOT: "~" ...
    L_LESS_THAN = "LESS_THAN"  # LESS_THAN: "<" ...
    L_GREATER_THAN = "GREATER_THAN"  # GREATER_THAN: ">" ...
    L_EQUAL_TO = "EQUAL_TO"  # EQUAL_TO: "==" ...
    L_NOT_EQUAL_TO = "NOT_EQUAL_TO"  # NOT_EQUAL_TO: "!=" ...
    L_GREATER_THAN_OR_EQUAL_TO = (
        "GREATER_THAN_OR_EQUAL_TO"  # GREATER_THAN_OR_EQUAL_TO: ">=" ...
    )
    L_LESS_THAN_OR_EQUAL_TO = "LESS_THAN_OR_EQUAL_TO"  # LESS_THAN_OR_EQUAL_TO: "<=" ...
    L_DOUBLE_STAR = "DOUBLE_STAR"  # DOUBLE_STAR: "**" ...
    L_DOUBLE_DIVIDE = "DOUBLE_DIVIDE"  # DOUBLE_DIVIDE: "//" ...
    L_SHIFT_LEFT = "SHIFT_LEFT"  # SHIFT_LEFT: "<<" ...
    L_SHIFT_RIGHT = "SHIFT_RIGHT"  # SHIFT_RIGHT: ">>" ...
    L_AT = "AT"  # AT: "@" ...
    L_COMP_PLUS = "COMP_PLUS"  # COMP_PLUS: "+=" ...
    L_COMP_MINUS = "COMP_MINUS"  # COMP_MINUS: "-=" ...
    L_COMP_MULTIPLY = "COMP_MULTIPLY"  # COMP_MULTIPLY: "*=" ...
    L_COMP_DIVIDE = "COMP_DIVIDE"  # COMP_DIVIDE: "/=" ...
    L_COMP_MOD = "COMP_MOD"  # COMP_MOD: "%=" ...
    L_COMP_AND = "COMP_AND"  # COMP_AND: "&=" ...
    L_COMP_OR = "COMP_OR"  # COMP_OR: "|=" ...
    L_COMP_XOR = "COMP_XOR"  # COMP_XOR: "^=" ...
    L_COMP_DOUBLE_STAR = "COMP_DOUBLE_STAR"  # COMP_DOUBLE_STAR: "**=" ...
    L_COMP_DOUBLE_DIVIDE = "COMP_DOUBLE_DIVIDE"  # COMP_DOUBLE_DIVIDE: "//=" ...
    L_COMP_SHIFT_LEFT = "COMP_SHIFT_LEFT"  # COMP_SHIFT_LEFT: "<<=" ...
    L_COMP_SHIFT_RIGHT = "COMP_SHIFT_RIGHT"  # COMP_SHIFT_RIGHT: ">>=" ...
    L_IMPORT = "IMPORT"  # IMPORT: "import" ...
    L_AS = "AS"  # AS: "as" ...
    L_RULE = "RULE"  # RULE: "rule" ...
    L_SCHEMA = "SCHEMA"  # SCHEMA: "schema" ...
    L_MIXIN = "MIXIN"  # MIXIN: "mixin" ...
    L_PROTOCOL = "PROTOCOL"  # PROTOCOL: "protocol" ...
    L_CHECK = "CHECK"  # CHECK: "check" ...
    L_FOR = "FOR"  # FOR: "for" ...
    L_ASSERT = "ASSERT"  # ASSERT: "assert" ...
    L_IF = "IF"  # IF: "if" ...
    L_ELIF = "ELIF"  # ELIF: "elif" ...
    L_ELSE = "ELSE"  # ELSE: "else" ...
    L_L_OR = "L_OR"  # L_OR: "or" ...
    L_L_AND = "L_AND"  # L_AND: "and" ...
    L_L_NOT = "L_NOT"  # L_NOT: "not" ...
    L_IN = "IN"  # IN: "in" ...
    L_IS = "IS"  # IS: "is" ...
    L_LAMBDA = "LAMBDA"  # LAMBDA: "lambda" ...
    L_ALL = "ALL"  # ALL: "all" ...
    L_ANY = "ANY"  # ANY: "any" ...
    L_FILTER = "FILTER"  # FILTER: "filter" ...
    L_MAP = "MAP"  # MAP: "map" ...
    L_TYPE = "TYPE"  # TYPE: "type" ...
    L_ANY_TYPE = "ANY_TYPE"  # ANY_TYPE: "any" ...
    L_STRING_TYPE = "STRING_TYPE"  # STRING_TYPE: "str" ...
    L_INT_TYPE = "INT_TYPE"  # INT_TYPE: "int" ...
    L_FLOAT_TYPE = "FLOAT_TYPE"  # FLOAT_TYPE: "float" ...
    L_BOOL_TYPE = "BOOL_TYPE"  # BOOL_TYPE: "bool" ...
    L_TRUE = "TRUE"  # TRUE: "True" ...
    L_FALSE = "FALSE"  # FALSE: "False" ...
    L_NONE = "NONE"  # NONE: "None" ...
    L_UNDEFINED = "UNDEFINED"  # UNDEFINED: "Undefined" ...
    L_SI_N_L = "SI_N_L"  # SI_N_L: "n" ...
    L_SI_U_L = "SI_U_L"  # SI_U_L: "u" ...
    L_SI_M_L = "SI_M_L"  # SI_M_L: "m" ...
    L_SI_K_L = "SI_K_L"  # SI_K_L: "k" ...
    L_SI_K = "SI_K"  # SI_K: "K" ...
    L_SI_M = "SI_M"  # SI_M: "M" ...
    L_SI_G = "SI_G"  # SI_G: "G" ...
    L_SI_T = "SI_T"  # SI_T: "T" ...
    L_SI_P = "SI_P"  # SI_P: "P" ...
    L_SI_K_IEC = "SI_K_IEC"  # SI_K_IEC: "Ki" ...
    L_SI_M_IEC = "SI_M_IEC"  # SI_M_IEC: "Mi" ...
    L_SI_G_IEC = "SI_G_IEC"  # SI_G_IEC: "Gi" ...
    L_SI_T_IEC = "SI_T_IEC"  # SI_T_IEC: "Ti" ...
    L_SI_P_IEC = "SI_P_IEC"  # SI_P_IEC: "Pi" ...
    L_IEC = "IEC"  # IEC: "i" ...
    L_NAME = "NAME"  # NAME: /\$?[a-zA-Z_]\w*/ ...
    L_COMMENT = "COMMENT"  # COMMENT: /#[^\n]*/ ...
    L_NEWLINE = "NEWLINE"  # NEWLINE: ( /\r?\n[\t ]*/ | COMMENT )+ ...
    L_STRING = "STRING"  # STRING: /r?("(?!"").*?(?<!\\)(\\\\)*?"|'(?!'').*?(?<!\\)(\\\\)*?')/i ...
    L_LONG_STRING = "LONG_STRING"  # LONG_STRING: /r?(""".*?(?<!\\)(\\\\)*?"""|'''.*?(?<!\\)(\\\\)*?''')/is ...
    L_DEC_NUMBER = "DEC_NUMBER"  # DEC_NUMBER: /\-?0|\-?[1-9]\d*/i ...
    L_HEX_NUMBER = "HEX_NUMBER"  # HEX_NUMBER.2: /\-?0[xX][0-9a-fA-F]+/i ...
    L_OCT_NUMBER = "OCT_NUMBER"  # OCT_NUMBER.2: /\-?0[oO]?[0-7]+/i ...
    L_BIN_NUMBER = "BIN_NUMBER"  # BIN_NUMBER.2 : /\-?0[bB][0-1]+/i ...
    L_FLOAT_NUMBER = "FLOAT_NUMBER"  # FLOAT_NUMBER.2: /(([-+]?\d+\.\d*|\.\d+)(e[-+]?\d+)?|\d+(e[-+]?\d+))/i ...

    # kcl.lark tokens list (len=103)
    LL_token_list = [
        L_ASSIGN,
        L_COLON,
        L_SEMI_COLON,
        L_COMMA,
        L_QUESTION,
        L_ELLIPSIS,
        L_RIGHT_ARROW,
        L_LEFT_PARENTHESES,
        L_RIGHT_PARENTHESES,
        L_LEFT_BRACKETS,
        L_RIGHT_BRACKETS,
        L_LEFT_BRACE,
        L_RIGHT_BRACE,
        L_PLUS,
        L_MINUS,
        L_MULTIPLY,
        L_DIVIDE,
        L_MOD,
        L_DOT,
        L_AND,
        L_OR,
        L_XOR,
        L_NOT,
        L_LESS_THAN,
        L_GREATER_THAN,
        L_EQUAL_TO,
        L_NOT_EQUAL_TO,
        L_GREATER_THAN_OR_EQUAL_TO,
        L_LESS_THAN_OR_EQUAL_TO,
        L_DOUBLE_STAR,
        L_DOUBLE_DIVIDE,
        L_SHIFT_LEFT,
        L_SHIFT_RIGHT,
        L_AT,
        L_COMP_PLUS,
        L_COMP_MINUS,
        L_COMP_MULTIPLY,
        L_COMP_DIVIDE,
        L_COMP_MOD,
        L_COMP_AND,
        L_COMP_OR,
        L_COMP_XOR,
        L_COMP_DOUBLE_STAR,
        L_COMP_DOUBLE_DIVIDE,
        L_COMP_SHIFT_LEFT,
        L_COMP_SHIFT_RIGHT,
        L_IMPORT,
        L_AS,
        L_RULE,
        L_SCHEMA,
        L_MIXIN,
        L_PROTOCOL,
        L_CHECK,
        L_FOR,
        L_ASSERT,
        L_IF,
        L_ELIF,
        L_ELSE,
        L_L_OR,
        L_L_AND,
        L_L_NOT,
        L_IN,
        L_IS,
        L_LAMBDA,
        L_ALL,
        L_ANY,
        L_FILTER,
        L_MAP,
        L_TYPE,
        L_ANY_TYPE,
        L_STRING_TYPE,
        L_INT_TYPE,
        L_FLOAT_TYPE,
        L_BOOL_TYPE,
        L_TRUE,
        L_FALSE,
        L_NONE,
        L_UNDEFINED,
        L_SI_N_L,
        L_SI_U_L,
        L_SI_M_L,
        L_SI_K_L,
        L_SI_K,
        L_SI_M,
        L_SI_G,
        L_SI_T,
        L_SI_P,
        L_SI_K_IEC,
        L_SI_M_IEC,
        L_SI_G_IEC,
        L_SI_T_IEC,
        L_SI_P_IEC,
        L_IEC,
        L_NAME,
        L_COMMENT,
        L_NEWLINE,
        L_STRING,
        L_LONG_STRING,
        L_DEC_NUMBER,
        L_HEX_NUMBER,
        L_OCT_NUMBER,
        L_BIN_NUMBER,
        L_FLOAT_NUMBER,
    ]

    # kcl.lark rules list (len=83)
    LL_rule_list = [
        L_start,
        L_statement,
        L_simple_stmt,
        L_compound_stmt,
        L_import_stmt,
        L_dot_name,
        L_leading_dots,
        L_assert_stmt,
        L_if_stmt,
        L_execution_block,
        L_if_simple_stmt,
        L_assign_stmt,
        L_simple_assign_stmt,
        L_unification_stmt,
        L_schema_stmt,
        L_schema_arguments,
        L_schema_argument,
        L_schema_body,
        L_schema_attribute_stmt,
        L_attribute_stmt,
        L_schema_init_stmt,
        L_schema_index_signature,
        L_rule_stmt,
        L_rule_body,
        L_for_host,
        L_decorators,
        L_decorator_expr,
        L_type,
        L_type_element,
        L_schema_type,
        L_basic_type,
        L_compound_type,
        L_list_type,
        L_dict_type,
        L_literal_type,
        L_type_alias_stmt,
        L_check_block,
        L_check_expr,
        L_mixin_stmt,
        L_multiline_mixins,
        L_mixins,
        L_expr_stmt,
        L_testlist_expr,
        L_test,
        L_if_expr,
        L_simple_expr,
        L_unary_expr,
        L_binary_expr,
        L_bin_op,
        L_un_op,
        L_primary_expr,
        L_operand,
        L_select_suffix,
        L_call_suffix,
        L_slice_suffix,
        L_arguments,
        L_argument,
        L_identifier,
        L_quant_expr,
        L_quant_target,
        L_quant_op,
        L_list_expr,
        L_list_items,
        L_list_item,
        L_list_comp,
        L_dict_comp,
        L_entry,
        L_comp_clause,
        L_if_entry,
        L_if_entry_exec_block,
        L_if_item,
        L_if_item_exec_block,
        L_star_expr,
        L_double_star_expr,
        L_loop_variables,
        L_schema_expr,
        L_config_expr,
        L_config_entries,
        L_config_entry,
        L_lambda_expr,
        L_number,
        L_multiplier,
        L_string,
    ]

    # kcl.lark tokens string value map
    LL_token_str_value_map = {
        L_ASSIGN: "=",
        L_COLON: ":",
        L_SEMI_COLON: ";",
        L_COMMA: ",",
        L_QUESTION: "?",
        L_ELLIPSIS: "...",
        L_RIGHT_ARROW: "->",
        L_LEFT_PARENTHESES: "(",
        L_RIGHT_PARENTHESES: ")",
        L_LEFT_BRACKETS: "[",
        L_RIGHT_BRACKETS: "]",
        L_LEFT_BRACE: "{",
        L_RIGHT_BRACE: "}",
        L_PLUS: "+",
        L_MINUS: "-",
        L_MULTIPLY: "*",
        L_DIVIDE: "/",
        L_MOD: "%",
        L_DOT: ".",
        L_AND: "&",
        L_OR: "|",
        L_XOR: "^",
        L_NOT: "~",
        L_LESS_THAN: "<",
        L_GREATER_THAN: ">",
        L_EQUAL_TO: "==",
        L_NOT_EQUAL_TO: "!=",
        L_GREATER_THAN_OR_EQUAL_TO: ">=",
        L_LESS_THAN_OR_EQUAL_TO: "<=",
        L_DOUBLE_STAR: "**",
        L_DOUBLE_DIVIDE: "//",
        L_SHIFT_LEFT: "<<",
        L_SHIFT_RIGHT: ">>",
        L_AT: "@",
        L_COMP_PLUS: "+=",
        L_COMP_MINUS: "-=",
        L_COMP_MULTIPLY: "*=",
        L_COMP_DIVIDE: "/=",
        L_COMP_MOD: "%=",
        L_COMP_AND: "&=",
        L_COMP_OR: "|=",
        L_COMP_XOR: "^=",
        L_COMP_DOUBLE_STAR: "**=",
        L_COMP_DOUBLE_DIVIDE: "//=",
        L_COMP_SHIFT_LEFT: "<<=",
        L_COMP_SHIFT_RIGHT: ">>=",
        L_IMPORT: "import",
        L_AS: "as",
        L_RULE: "rule",
        L_SCHEMA: "schema",
        L_MIXIN: "mixin",
        L_PROTOCOL: "protocol",
        L_CHECK: "check",
        L_FOR: "for",
        L_ASSERT: "assert",
        L_IF: "if",
        L_ELIF: "elif",
        L_ELSE: "else",
        L_L_OR: "or",
        L_L_AND: "and",
        L_L_NOT: "not",
        L_IN: "in",
        L_IS: "is",
        L_LAMBDA: "lambda",
        L_ALL: "all",
        L_ANY: "any",
        L_FILTER: "filter",
        L_MAP: "map",
        L_TYPE: "type",
        L_ANY_TYPE: "any",
        L_STRING_TYPE: "str",
        L_INT_TYPE: "int",
        L_FLOAT_TYPE: "float",
        L_BOOL_TYPE: "bool",
        L_TRUE: "True",
        L_FALSE: "False",
        L_NONE: "None",
        L_UNDEFINED: "Undefined",
        L_SI_N_L: "n",
        L_SI_U_L: "u",
        L_SI_M_L: "m",
        L_SI_K_L: "k",
        L_SI_K: "K",
        L_SI_M: "M",
        L_SI_G: "G",
        L_SI_T: "T",
        L_SI_P: "P",
        L_SI_K_IEC: "Ki",
        L_SI_M_IEC: "Mi",
        L_SI_G_IEC: "Gi",
        L_SI_T_IEC: "Ti",
        L_SI_P_IEC: "Pi",
        L_IEC: "i",
    }


class TokenValue:
    ASSIGN = "="
    COLON = ":"
    SEMI_COLON = ";"
    COMMA = ","
    QUESTION = "?"
    ELLIPSIS = "..."
    RIGHT_ARROW = "->"
    LEFT_PARENTHESES = "("
    RIGHT_PARENTHESES = ")"
    LEFT_BRACKETS = "["
    RIGHT_BRACKETS = "]"
    LEFT_BRACE = "{"
    RIGHT_BRACE = "}"
    PLUS = "+"
    MINUS = "-"
    MULTIPLY = "*"
    DIVIDE = "/"
    MOD = "%"
    DOT = "."
    AND = "&"
    OR = "|"
    XOR = "^"
    NOT = "~"
    LESS_THAN = "<"
    GREATER_THAN = ">"
    EQUAL_TO = "=="
    NOT_EQUAL_TO = "!="
    GREATER_THAN_OR_EQUAL_TO = ">="
    LESS_THAN_OR_EQUAL_TO = "<="
    DOUBLE_STAR = "**"
    DOUBLE_DIVIDE = "//"
    SHIFT_LEFT = "<<"
    SHIFT_RIGHT = ">>"
    AT = "@"
    COMP_PLUS = "+="
    COMP_MINUS = "-="
    COMP_MULTIPLY = "*="
    COMP_DIVIDE = "/="
    COMP_MOD = "%="
    COMP_AND = "&="
    COMP_OR = "|="
    COMP_XOR = "^="
    COMP_DOUBLE_STAR = "**="
    COMP_DOUBLE_DIVIDE = "//="
    COMP_SHIFT_LEFT = "<<="
    COMP_SHIFT_RIGHT = ">>="
    IMPORT = "import"
    AS = "as"
    RULE = "rule"
    SCHEMA = "schema"
    MIXIN = "mixin"
    PROTOCOL = "protocol"
    CHECK = "check"
    FOR = "for"
    ASSERT = "assert"
    IF = "if"
    ELIF = "elif"
    ELSE = "else"
    L_OR = "or"
    L_AND = "and"
    L_NOT = "not"
    IN = "in"
    IS = "is"
    LAMBDA = "lambda"
    ALL = "all"
    ANY = "any"
    FILTER = "filter"
    MAP = "map"
    TYPE = "type"
    ANY_TYPE = "any"
    STRING_TYPE = "str"
    INT_TYPE = "int"
    FLOAT_TYPE = "float"
    BOOL_TYPE = "bool"
    TRUE = "True"
    FALSE = "False"
    NONE = "None"
    UNDEFINED = "Undefined"
    SI_N_L = "n"
    SI_U_L = "u"
    SI_M_L = "m"
    SI_K_L = "k"
    SI_K = "K"
    SI_M = "M"
    SI_G = "G"
    SI_T = "T"
    SI_P = "P"
    SI_K_IEC = "Ki"
    SI_M_IEC = "Mi"
    SI_G_IEC = "Gi"
    SI_T_IEC = "Ti"
    SI_P_IEC = "Pi"
    IEC = "i"



import typing
import pygls.lsp.types.language_features.hover as pygls_hover
from pygls.lsp.types.basic_structures import MarkupContent, MarkupKind, Range
import kclvm.kcl.ast as ast
import kclvm.api.object as objpkg
from kclvm.kcl.types.scope import ScopeObject
import kclvm.tools.langserver.common as common
from . import go_to_def


def hover(pos: ast.Position, code: str = None) -> typing.Optional[pygls_hover.Hover]:
    def_node, def_obj = go_to_def.definition(pos, code)
    if not def_node:
        return None
    if def_obj:
        return scope_obj_desc(def_node, def_obj)
    else:
        return ast_node_desc(def_node)


def ast_node_desc(node: ast.AST) -> typing.Optional[pygls_hover.Hover]:
    if isinstance(node, ast.Name):
        return pygls_hover.Hover(
            contents=MarkupContent(
                kind=MarkupKind.PlainText,
                value=node.value,
            ),
            range=Range(
                start=common.kcl_pos_to_lsp_pos(node.pos),
                end=common.kcl_pos_to_lsp_pos(node.end_pos),
            ),
        )
    return None


def scope_obj_desc(
    node: ast.AST, obj: ScopeObject
) -> typing.Optional[pygls_hover.Hover]:
    if isinstance(node, ast.AST) and isinstance(obj, ScopeObject):
        # 针对每种类型的 scope object，进行不同的显示
        if not obj.node and isinstance(obj.type, objpkg.KCLFunctionTypeObject):
            # the target scope object is a built-in function name
            msg = (
                f"(built-in) {obj.name}("
                + ", ".join([param.param_doc() for param in obj.type.params])
                + f"): {obj.type.return_type.type_str() if obj.type.return_type else 'any'}"
                + f"\n{obj.type.doc}"
                if obj.type.doc
                else ""
            )
        else:
            msg = (
                f"{obj.name}\ntype: {obj.type.type_str()}\ndefined in:{obj.node.filename}"
                if obj.node
                else obj.name
            )
        return pygls_hover.Hover(
            contents=MarkupContent(
                kind=MarkupKind.PlainText,
                value=msg,
            ),
            range=Range(
                start=common.kcl_pos_to_lsp_pos(node.pos),
                end=common.kcl_pos_to_lsp_pos(node.end_pos),
            ),
        )
    return None
